/*
 * AutomataUnionCartesianProduct.h
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_UNION_CARTESIAN_H_
#define AUTOMATA_UNION_CARTESIAN_H_

#include <core/multipleDispatch.hpp>
#include <automaton/Automaton.h>
#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Union two automata.
 *  - For finite automata A1, A2, we create automaton L accepting L(A1) \cup L(A2) (Melichar, 2.71)
 */
class AutomataUnionCartesianProduct : public alib::PromotingDoubleDispatch<AutomataUnionCartesianProduct, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	static automaton::Automaton unification(const automaton::Automaton& first, const automaton::Automaton& second);

	template < class SymbolType, class StateType1, class StateType2 >
	static automaton::NFA < SymbolType, std::pair < StateType1, StateType2 > > unification(const automaton::NFA < SymbolType, StateType1 > & first, const automaton::NFA < SymbolType, StateType2 > & second);
	template < class SymbolType, class StateType1, class StateType2 >
	static automaton::DFA < SymbolType, std::pair < StateType1, StateType2 > > unification(const automaton::DFA < SymbolType, StateType1 > & first, const automaton::DFA < SymbolType, StateType2 > & second);
};

template < class SymbolType, class StateType1, class StateType2 >
automaton::DFA < SymbolType, std::pair < StateType1, StateType2 > > AutomataUnionCartesianProduct::unification(const automaton::DFA < SymbolType, StateType1 > & first, const automaton::DFA < SymbolType, StateType2 > & second) {
	if(!first.isTotal() || !second.isTotal())
		throw exception::CommonException("Automata must be total to unify with cartesian product");

	std::pair < StateType1, StateType2 > q0 ( first.getInitialState(), second.getInitialState() );
	automaton::DFA < SymbolType, std::pair < StateType1, StateType2 > > res ( q0 );

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(std::make_pair(p, q));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getStates())
			res.addFinalState(std::make_pair(p, q));

	for(const auto& p : first.getStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(std::make_pair(p, q));

	for(const auto& state : res.getStates())
		for(const auto& tp : first.getTransitionsFromState(state.first))
			for(const auto& tq : second.getTransitionsFromState(state.second))
				if(tp.first.second == tq.first.second)
					res.addTransition(state, tp.first.second, std::make_pair(tp.second, tq.second));


	return res;
}

template < class SymbolType, class StateType1, class StateType2 >
automaton::NFA < SymbolType, std::pair < StateType1, StateType2 > > AutomataUnionCartesianProduct::unification(const automaton::NFA < SymbolType, StateType1 > & first, const automaton::NFA < SymbolType, StateType2 > & second) {
	if(!first.isTotal() || !second.isTotal())
		throw exception::CommonException("Automata must be total to unify with cartesian product");

	std::pair < StateType1, StateType2 > q0 ( first.getInitialState(), second.getInitialState() );
	automaton::NFA < SymbolType, std::pair < StateType1, StateType2 > > res(q0);

	for(const auto& a : first.getInputAlphabet())
		res.addInputSymbol(a);
	for(const auto& a : second.getInputAlphabet())
		res.addInputSymbol(a);

	for(const auto& p : first.getStates())
		for(const auto& q : second.getStates())
			res.addState(std::make_pair(p, q));

	for(const auto& p : first.getFinalStates())
		for(const auto& q : second.getStates())
			res.addFinalState(std::make_pair(p, q));

	for(const auto& p : first.getStates())
		for(const auto& q : second.getFinalStates())
			res.addFinalState(std::make_pair(p, q));

	for(const auto& state : res.getStates())
		for(const auto& tp : first.getTransitionsFromState(state.first))
			for(const auto& tq : second.getTransitionsFromState(state.second))
				if(tp.first.second == tq.first.second)
					for(const auto& p : tp.second)
						for(const auto& q : tq.second)
							res.addTransition(state, tp.first.second, std::make_pair(p, q));

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_UNION_CARTESIAN_H_ */
