/*
 * EpsilonRemover.h
 *
 *  Created on: 24. 11. 2014
 *	  Author: Jan Travnicek
 */

#ifndef EPSILON_REMOVER_H_
#define EPSILON_REMOVER_H_

#include <core/multipleDispatch.hpp>
#include <map>
#include <algorithm>

#include <grammar/Grammar.h>

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>

#include "../properties/NullableNonterminals.h"

namespace grammar {

namespace simplify {

class EpsilonRemover : public alib::SingleDispatch<EpsilonRemover, grammar::Grammar, const grammar::GrammarBase &> {
	template < class SymbolType >
	static void removeNullableNonterminals(grammar::EpsilonFreeCFG < SymbolType > & grammar, const std::set<SymbolType>& nullableNonterminals, const SymbolType& lhs, const std::vector<SymbolType>& rhs, unsigned i, std::vector<SymbolType> clear);

	template<class T, class SymbolType = typename grammar::SymbolTypeOfGrammar < T > >
	static grammar::EpsilonFreeCFG < SymbolType > removeInternal( const T & origGrammar );
public:
	static grammar::Grammar remove( const grammar::Grammar & grammar );

	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::CFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::EpsilonFreeCFG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::GNF < SymbolType > remove( const grammar::GNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::CNF < SymbolType > remove( const grammar::CNF < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::LG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::LeftLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::LeftRG < SymbolType > remove( const grammar::LeftRG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::EpsilonFreeCFG < SymbolType > remove( const grammar::RightLG < SymbolType > & grammar );
	template < class SymbolType >
	static grammar::RightRG < SymbolType > remove( const grammar::RightRG < SymbolType > & grammar );
};

template < class SymbolType >
void EpsilonRemover::removeNullableNonterminals(grammar::EpsilonFreeCFG < SymbolType > & grammar, const std::set<SymbolType>& nullableNonterminals, const SymbolType& lhs, const std::vector<SymbolType>& rhs, unsigned i, std::vector<SymbolType> clear) {
	if(rhs.size() == i) {
		if(clear.size() == 0) return;
		grammar.addRule(lhs, clear);
	} else if(nullableNonterminals.find(rhs[i]) != nullableNonterminals.end()) {
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
		clear.push_back(rhs[i]);
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
	} else {
		clear.push_back(rhs[i]);
		removeNullableNonterminals(grammar, nullableNonterminals, lhs, rhs, i+1, clear);
	}
}

template<class T, class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::removeInternal( const T & origGrammar ) {
	grammar::EpsilonFreeCFG < SymbolType > grammar(origGrammar.getInitialSymbol());

	for( const auto & symbol : origGrammar.getNonterminalAlphabet() )
		grammar.addNonterminalSymbol( symbol );

	for( const auto & symbol : origGrammar.getTerminalAlphabet() )
		grammar.addTerminalSymbol( symbol );

	std::set<SymbolType> nullableNonterminals = grammar::properties::NullableNonterminals::getNullableNonterminals(origGrammar);

	for( const auto & rule : origGrammar.getRawRules( ) )
		for(const auto & rhs : rule.second) {
			if(rhs.size() == 0) continue;

			EpsilonRemover::removeNullableNonterminals(grammar, nullableNonterminals, rule.first, rhs, 0, {});
		}

	if(nullableNonterminals.find(origGrammar.getInitialSymbol()) != nullableNonterminals.end())
		grammar.setGeneratesEpsilon(true);

	return grammar;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::remove(const grammar::CFG < SymbolType > & origGrammar) {
	return EpsilonRemover::removeInternal(origGrammar);
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::remove(const grammar::EpsilonFreeCFG < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::CNF < SymbolType > EpsilonRemover::remove(const grammar::CNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::GNF < SymbolType > EpsilonRemover::remove(const grammar::GNF < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::remove(const grammar::LG < SymbolType > & origGrammar) {
	return EpsilonRemover::removeInternal(origGrammar);
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::remove(const grammar::LeftLG < SymbolType > & origGrammar) {
	return EpsilonRemover::removeInternal(origGrammar);
}

template < class SymbolType >
grammar::LeftRG < SymbolType > EpsilonRemover::remove(const grammar::LeftRG < SymbolType > & origGrammar) {
	return origGrammar;
}

template < class SymbolType >
grammar::EpsilonFreeCFG < SymbolType > EpsilonRemover::remove(const grammar::RightLG < SymbolType > & origGrammar) {
	return EpsilonRemover::removeInternal(origGrammar);
}

template < class SymbolType >
grammar::RightRG < SymbolType > EpsilonRemover::remove(const grammar::RightRG < SymbolType > & origGrammar) {
	return origGrammar;
}

} /* namespace simplify */

} /* namespace grammar */

#endif /* EPSILON_REMOVER_H_ */
