/*
 * SubtreeJumpTable.h
 *
 *  Created on: 5. 11. 2014
 *      Author: Jan Travnicek
 */

#ifndef _SUBTREE_JUMP_TABLE_H_
#define _SUBTREE_JUMP_TABLE_H_

#include <vector>

#include <tree/RankedTreeWrapper.h>
#include <tree/TreeFeatures.h>
#include <core/multipleDispatch.hpp>

#include <tree/ranked/PrefixRankedTree.h>
#include <tree/ranked/PrefixRankedPattern.h>
#include <tree/ranked/PrefixRankedBarTree.h>
#include <tree/ranked/PrefixRankedBarPattern.h>

namespace tree {

namespace properties {

class SubtreeJumpTable : public alib::SingleDispatch < SubtreeJumpTable, std::vector < int >, const tree::RankedTreeBase & > {
	template < class T >
	static int buildDataPointersBar ( std::vector < int > & res, const T & subject, int begin );
	template < class T >
	static int buildDataPointersPrefixRanked ( std::vector < int > & res, const T & subject, int begin );
	template < class T >
	static int buildDataPointersPrefixRankedInternal ( std::vector < int > & res, const T & subject, int begin );

public:
	static std::vector < int > compute ( const tree::RankedTreeWrapper & subject );

	template < class SymbolType, class RankType >
	static std::vector < int > compute ( const tree::PrefixRankedBarTree < SymbolType, RankType > & subject );
	template < class SymbolType, class RankType >
	static std::vector < int > compute ( const tree::PrefixRankedBarPattern < SymbolType, RankType > & pattern );
	template < class SymbolType, class RankType >
	static std::vector < int > compute ( const tree::PrefixRankedTree < SymbolType, RankType > & subject );
	template < class SymbolType, class RankType >
	static std::vector < int > compute ( const tree::PrefixRankedPattern < SymbolType, RankType > & pattern );

};

template < class SymbolType, class RankType >
std::vector < int > SubtreeJumpTable::compute ( const tree::PrefixRankedBarTree < SymbolType, RankType > & subject ) {
	std::vector < int > res;

	buildDataPointersBar ( res, subject, 0 );

	return res;
}

template < class SymbolType, class RankType >
std::vector < int > SubtreeJumpTable::compute ( const tree::PrefixRankedBarPattern < SymbolType, RankType > & pattern ) {
	std::vector < int > res;

	buildDataPointersBar ( res, pattern, 0 );

	return res;
}

template < class SymbolType, class RankType >
std::vector < int > SubtreeJumpTable::compute ( const tree::PrefixRankedTree < SymbolType, RankType > & subject ) {
	std::vector < int > res;

	buildDataPointersPrefixRanked ( res, subject, 0 );

	return res;
}

template < class SymbolType, class RankType >
std::vector < int > SubtreeJumpTable::compute ( const tree::PrefixRankedPattern < SymbolType, RankType > & pattern ) {
	std::vector < int > res;

	buildDataPointersPrefixRanked ( res, pattern, 0 );

	return res;
}

/**
 * used to compute subtree jump table.
 * @param begin - index of a root node of a complete subtree to process
 * @return index, increased by one, of the last node in the subtree starting at index begin
 */
template < class T >
int SubtreeJumpTable::buildDataPointersBar ( std::vector < int > & res, const T & subject, int begin ) {
	res.push_back ( 0 );
	int index = begin + 1;

	if ( ! subject.getBars ( ).count ( subject.getContent ( )[begin] ) )
		for ( unsigned i = 0; i < ( unsigned ) subject.getContent ( )[begin].getRank ( ); i++ )
			index = buildDataPointersBar ( res, subject, index );

	index++;
	res[begin] = index;
	res.push_back ( begin - 1 );
	return index;
}

/**
 * used to compute subtree jump table.
 * @param begin - index of a root node of a complete subtree to process
 * @return index, increased by one, of the last node in the subtree starting at index begin
 */
template < class T >
int SubtreeJumpTable::buildDataPointersPrefixRanked ( std::vector < int > & res, const T & subject, int begin ) {
	for ( unsigned i = 0; i < subject.getContent ( ).size ( ); i++ )
		res.push_back ( 0 );

	return buildDataPointersPrefixRankedInternal ( res, subject, begin );
}

template < class T >
int SubtreeJumpTable::buildDataPointersPrefixRankedInternal ( std::vector < int > & res, const T & subject, int begin ) {
	int index = begin + 1;

	for ( unsigned i = 0; i < ( unsigned ) subject.getContent ( )[begin].getRank ( ); i++ )
		index = buildDataPointersPrefixRankedInternal ( res, subject, index );

	res[begin] = index;
	return index;
}

} /* namespace properties */

} /* namespace tree */

#endif /* _SUBTREE_JUMP_TABLE_H_ */
