/*
 * First.cpp
 *
 *  Created on: 9. 6. 2015
 *	  Author: Tomas Pecka
 */

#include "First.h"

#include <grammar/ContextFree/CFG.h>
#include <grammar/ContextFree/EpsilonFreeCFG.h>
#include <grammar/ContextFree/CNF.h>
#include <grammar/ContextFree/GNF.h>
#include <grammar/ContextFree/LG.h>
#include <grammar/Regular/LeftLG.h>
#include <grammar/Regular/LeftRG.h>
#include <grammar/Regular/RightLG.h>
#include <grammar/Regular/RightRG.h>
#include <exception/CommonException.h>
#include <registration/AlgoRegistration.hpp>

namespace grammar {

namespace parsing {

std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > First::first ( const std::set < DefaultSymbolType > & terminals, const std::set < DefaultSymbolType > & nonterminals, const std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > & firstOfNonterminal, const std::vector < DefaultSymbolType > & rhs ) {
	 // 1. FIRST(\varepsilon) = { \varepsilon }
	if ( rhs.size ( ) == 0 ) {
		return { string::Epsilon < >::EPSILON };
	}

	 // 2. FIRST(a) = { a } forall a \in T
	else if ( terminals.count ( rhs[0] ) ) {
		return { rhs[0] };
	}

	 // 4. FIRST(A \alpha) = first(A) if A \in N and \varepsilon \notin first(A)
	else if ( nonterminals.count ( rhs[0] ) && !firstOfNonterminal.find ( rhs[0] )->second.count ( string::Epsilon < >::EPSILON ) ) {
		return firstOfNonterminal.find ( rhs[0] )->second;
	}

	 // 5. FIRST(A \alpha) = (first(A) - \varepsilon) \cup FIRST(\alpha) if A \in N and \varepsilon \in first(A)
	else if ( nonterminals.count ( rhs[0] ) && firstOfNonterminal.find ( rhs[0] )->second.count ( string::Epsilon < >::EPSILON ) ) {
		std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > res = firstOfNonterminal.find ( rhs[0] )->second;
		res.erase ( string::Epsilon < >::EPSILON );

		std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > next = first ( terminals, nonterminals, firstOfNonterminal, std::vector < DefaultSymbolType > ( rhs.begin ( ) + 1, rhs.end ( ) ) );
		res.insert ( next.begin ( ), next.end ( ) );
		return res;
	} else {
		throw exception::CommonException ( "Cant be reached" );
	}
}

std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > First::first ( const std::set < DefaultSymbolType > & terminals, const std::set < DefaultSymbolType > & nonterminals, const std::map < DefaultSymbolType, std::set < std::vector < DefaultSymbolType > > > & rules ) {
	/*
	 *
	 * 1. foreach A \in N: first(A) = \emptyset
	 * 2. foreach A \rightarrow \alpha:
	 *	first(A) = first(A) \cup FIRST(\alpha)
	 * 3. repeat step 2 if at least one set first(A) has changed
	 *
	 */
	std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > firstOfNonterminal1;

	for ( const DefaultSymbolType & nonterminal : nonterminals )
		firstOfNonterminal1[nonterminal];

	std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > firstOfNonterminal2 = firstOfNonterminal1;

	do {
		for ( const std::pair < const DefaultSymbolType, std::set < std::vector < DefaultSymbolType > > > & rule : rules )
			for ( const std::vector < DefaultSymbolType > & rhs : rule.second ) {
				std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > newFirst = first ( terminals, nonterminals, firstOfNonterminal1, rhs );
				firstOfNonterminal2[rule.first].insert ( newFirst.begin ( ), newFirst.end ( ) );
			}

		if ( firstOfNonterminal1 == firstOfNonterminal2 ) break;

		firstOfNonterminal1 = std::move ( firstOfNonterminal2 );
		firstOfNonterminal2.clear ( );

	} while ( true );

	return firstOfNonterminal1;
}

template < class T >
std::map < std::vector < DefaultSymbolType >, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > First::first ( const T & grammar ) {
	std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > firstNt = first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), grammar.getRawRules ( ) );

	std::map < std::vector < DefaultSymbolType >, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > res;

	for ( const std::pair < const DefaultSymbolType, std::set < std::vector < DefaultSymbolType > > > & rule : grammar.getRawRules ( ) )
		for ( const std::vector < DefaultSymbolType > & rhs : rule.second )
			res.insert ( make_pair ( rhs, first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), firstNt, rhs ) ) );

	return res;
}

template < class T >
std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > First::first ( const T & grammar, const std::vector < DefaultSymbolType > & rhs ) {
	std::map < DefaultSymbolType, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > firstNt = first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), grammar.getRawRules ( ) );

	return first ( grammar.getTerminalAlphabet ( ), grammar.getNonterminalAlphabet ( ), firstNt, rhs );
}

auto FirstCFG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::CFG < > > ( First::first );
auto FirstEpsilonFreeCFG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::EpsilonFreeCFG < > > ( First::first );
auto FirstGNF = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::GNF < > > ( First::first );
auto FirstCNF = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::CNF < > > ( First::first );
auto FirstLG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::LG < > > ( First::first );
auto FirstLeftLG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::LeftLG < > > ( First::first );
auto FirstLeftRG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::LeftRG < > > ( First::first );
auto FirstRightLG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::RightLG < > > ( First::first );
auto FirstRightRG = registration::OverloadRegister < FirstBase1, FirstResult1, grammar::RightRG < > > ( First::first );

std::map < std::vector < DefaultSymbolType >, std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > > First::first ( const grammar::Grammar & grammar ) {
	return FirstBase1::dispatch ( grammar.getData ( ) );
}

auto FirstCFG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::CFG < > > ( First::first );
auto FirstEpsilonFreeCFG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::EpsilonFreeCFG < > > ( First::first );
auto FirstGNF2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::GNF < > > ( First::first );
auto FirstCNF2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::CNF < > > ( First::first );
auto FirstLG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::LG < > > ( First::first );
auto FirstLeftLG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::LeftLG < > > ( First::first );
auto FirstLeftRG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::LeftRG < > > ( First::first );
auto FirstRightLG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::RightLG < > > ( First::first );
auto FirstRightRG2 = registration::OverloadRegister < FirstBase2, FirstResult2, grammar::RightRG < > > ( First::first );

std::set < ext::variant < DefaultSymbolType, string::Epsilon < > > > First::first ( const grammar::Grammar & grammar, const std::vector < DefaultSymbolType > & rhs ) {
	return FirstBase2::dispatch ( grammar.getData ( ), rhs );
}

} /* namespace parsing */

} /* namespace grammar */
