/*
 * LR0Parser.cpp
 *
 *  Created on: 2. 5. 2016
 *	  Author: Martin Kocicka
 */

#include "LR0Parser.h"

#include "LRParser.h"

#include <label/LR0ItemsLabel.h>

#include <queue>

namespace grammar {

namespace parsing {
	
LR0Items LR0Parser::getClosure ( LR0Items items, grammar::CFG < > originalGrammar ) {
	bool changed;
	do {
		changed = false;

		for ( const LR0Items::value_type & symbolItems : items ) {
			for ( const std::pair < unsigned, std::vector < DefaultSymbolType > > & item : symbolItems.second ) {
				unsigned position = item.first;
				std::vector < DefaultSymbolType > rightHandSide = item.second;

				if ( position == rightHandSide.size ( ) )
					continue;

				std::map < DefaultSymbolType, std::set < std::vector < DefaultSymbolType > > >::const_iterator rulesIterator = originalGrammar.getRules ( ) . find(rightHandSide[position]);
				if ( rulesIterator == originalGrammar.getRules ( ) . end ( ) )
					continue;

				for ( const std::vector < DefaultSymbolType > & rule : rulesIterator->second ) {
					if (items[rightHandSide[position]].find ( { 0, rule } ) == items[rightHandSide[position]].end ( ) ) {
						changed = true;
						items[rightHandSide[position]].insert ( { 0, rule } );
					}
				}
			}
		}
	} while ( changed );

	return items;
}

LR0Items LR0Parser::getNextStateItems ( LR0Items items, DefaultSymbolType symbol, grammar::CFG < > originalGrammar ) {
	LR0Items transitionItems;
	for ( const LR0Items::value_type & symbolItems : items ) {
		DefaultSymbolType leftHandSide = symbolItems.first;
		for ( const std::pair < unsigned, std::vector < DefaultSymbolType > > & item : symbolItems.second) {
			unsigned position = item.first;
			std::vector < DefaultSymbolType > rightHandSide = item.second;

			if ( position == rightHandSide.size ( ) ) {
				continue;
			}

			if ( rightHandSide[position] == symbol ) {
				transitionItems[leftHandSide].insert ( { position + 1, rightHandSide } );
			}
		}
	}

	return getClosure ( transitionItems, originalGrammar );
}

automaton::DFA<> LR0Parser::getAutomaton ( grammar::CFG < > originalGrammar ) {
	grammar::CFG < > augmentedGrammar = LRParser::getAugmentedGrammar ( originalGrammar );
	DefaultSymbolType initialSymbol = augmentedGrammar.getInitialSymbol ( );
	const std::map < DefaultSymbolType, std::set < std::vector < DefaultSymbolType > > > & rules = augmentedGrammar.getRules ( );

	LR0Items initialItems;
	initialItems[initialSymbol].insert ( { 0, *rules.find ( initialSymbol ) -> second.begin ( ) } );

	DefaultStateType initialState ( label::LR0ItemsLabel ( getClosure ( initialItems, augmentedGrammar ) ) );

	automaton::DFA<> lr0Automaton ( initialState );
	lr0Automaton.addInputSymbols ( augmentedGrammar.getNonterminalAlphabet ( ) );
	lr0Automaton.addInputSymbols ( augmentedGrammar.getTerminalAlphabet ( ) );

	std::queue < DefaultStateType > itemsToProcess;
	itemsToProcess.push ( initialState );
	while ( !itemsToProcess.empty ( ) ) {
		DefaultStateType currentState = itemsToProcess.front ( );
		label::LR0ItemsLabel & currentLabel = static_cast < label::LR0ItemsLabel & > ( currentState . getData ( ) );
		itemsToProcess.pop ( );

		for ( const DefaultSymbolType & symbol : lr0Automaton.getInputAlphabet ( ) ) {
			LR0Items nextStateItems = getNextStateItems ( currentLabel.getItems ( ), symbol, augmentedGrammar );
			if ( !nextStateItems.empty ( ) ) {
				label::LR0ItemsLabel nextStateLabel ( nextStateItems );

				DefaultStateType nextState ( nextStateLabel );
				std::set < DefaultStateType > ::iterator stateIterator = lr0Automaton.getStates ( ) . find ( nextState );
				if ( stateIterator == lr0Automaton.getStates ( ) . end ( ) ) {
					lr0Automaton.addState ( nextState );
					lr0Automaton.addTransition ( currentState, symbol, nextState );
					itemsToProcess.push ( nextState );
				} else {
					lr0Automaton.addTransition ( currentState, symbol, *stateIterator );
				}
			}
		}
	}

	return lr0Automaton;
}

} /* namespace parsing */

} /* namespace grammar */
