/*
 * CastRegistry.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _CAST_REGISTRY_HPP_
#define _CAST_REGISTRY_HPP_

#include <memory>
#include <string>

#include <exception/CommonException.h>
#include <abstraction/OperationAbstraction.hpp>

namespace abstraction {

class CastRegistry {
	class Entry {
	public:
		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const = 0;

	};

	template < class Return, class Param >
	class DefaultEntryImpl : public Entry {
	public:
		DefaultEntryImpl ( ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	template < class Return, class Param >
	class AlgorithmEntryImpl : public Entry {
		std::function < Return ( Param ) > m_callback;
	public:
		AlgorithmEntryImpl ( std::function < Return ( Param ) > callback ) : m_callback ( callback ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	static std::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > & getEntries ( ) {
		static std::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > casts;
		return casts;
	};

public:
	template < class TargetType, class ParamType >
	static void registerCast ( std::string target, std::string param ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new DefaultEntryImpl < TargetType, ParamType > ( ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCast ( ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCast < TargetType, ParamType > ( target, param );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( std::string target, std::string param, TargetType ( * callback ) ( ParamType ) ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new AlgorithmEntryImpl < TargetType, ParamType > ( callback ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( TargetType ( * callback ) ( ParamType ) ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCastAlgorithm < TargetType, ParamType > ( target, param, callback );
	}

	static std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( std::string target, std::string param ) {
		auto cast = getEntries ( ).find ( std::make_pair ( target, param ) );
		if ( cast == getEntries ( ).end ( ) )
			throw exception::CommonException ( "Entry from " + param + " to " + target + " not available." );

		return cast->second->getAbstraction ( );
	}

};

} /* namespace abstraction */

#include <abstraction/CastAbstraction.hpp>
#include <abstraction/AlgorithmAbstraction.hpp>

namespace abstraction {

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::DefaultEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::CastAbstraction < Return, Param > > ( );
}

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::AlgorithmEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::AlgorithmAbstraction < Return, Param > > ( m_callback );
}

} /* namespace abstraction */

#endif /* _CAST_REGISTRY_HPP_ */
