/*
 * Deque.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_DEQUE_H_
#define OBJECTS_DEQUE_H_

#include <deque>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occurred.
 */
template < class ElementType >
class ObjectsDeque final : public std::deque < ElementType >, public ContainerBase {
public:
	explicit ObjectsDeque ( std::deque < ElementType > );
	explicit ObjectsDeque ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsDeque & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Deque";

		return xmlTagName;
	}

	static ObjectsDeque < ElementType > parse ( std::deque < sax::Token >::iterator & input );
	static std::deque < ElementType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::deque < ElementType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	static std::deque < alib::Object > normalizeRaw ( std::deque < ElementType > && source ) {
		std::deque < alib::Object > res;
		for ( ElementType & element : source ) {
			res.push_back ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( element ) ) ) );
		}
		return res;
	}

	typedef ObjectsDeque < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsDeque < alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( std::deque < ElementType > raw ) : std::deque < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( ) : std::deque < ElementType > ( ) {
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::clone ( ) const {
	return new ObjectsDeque ( * this );
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::plunder ( ) && {
	return new ObjectsDeque ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsDeque < ElementType >::compare ( const ObjectsDeque & other ) const {
	static ext::compare < std::deque < ElementType > > comp;

	return comp ( static_cast < const std::deque < ElementType > & > ( * this ), static_cast < const std::deque < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsDeque < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsDeque " << static_cast < const std::deque < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsDeque < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
ObjectsDeque < ElementType > ObjectsDeque < ElementType >::parse ( std::deque < sax::Token >::iterator & input ) {
	return ObjectsDeque < ElementType > ( parseRaw ( input ) );
}

template < class ElementType >
void ObjectsDeque < ElementType >::compose ( std::deque < sax::Token > & out ) const {
	compose ( out, * this );
}

template < class ElementType >
std::deque < ElementType > ObjectsDeque < ElementType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsDeque < >::getXmlTagName() );

	std::deque < ElementType > deque;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		deque.push_back ( alib::xmlApi < ElementType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsDeque < >::getXmlTagName() );
	return deque;
}

template < class ElementType >
void ObjectsDeque < ElementType >::compose ( std::deque < sax::Token > & out, const std::deque < ElementType > & input ) {
	out.emplace_back ( ObjectsDeque < >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	for ( const ElementType & item : input )
		alib::xmlApi < ElementType >::compose ( out, item );

	out.emplace_back ( ObjectsDeque < >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class ElementType >
alib::ObjectBase* ObjectsDeque < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < std::deque < T > > {
	static std::deque < T > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( std::deque < sax::Token > & output, const std::deque < T > & data );
};

template < typename T >
std::deque < T > xmlApi < std::deque < T > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsDeque < T >::parseRaw ( input );
}

template < typename T >
bool xmlApi < std::deque < T > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName() );
}

template < typename T >
std::string xmlApi < std::deque < T > >::xmlTagName ( ) {
	return container::ObjectsDeque < >::getXmlTagName();
}

template < typename T >
void xmlApi < std::deque < T > >::compose ( std::deque < sax::Token > & output, const std::deque < T > & input ) {
	return container::ObjectsDeque < T >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_DEQUE_H_ */
