/*
 * Pair.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_PAIR_H_
#define OBJECTS_PAIR_H_

#include <pair>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class FirstType, class SecondType >
class ObjectsPair final : public std::pair < FirstType, SecondType >, public ContainerBase {
public:
	explicit ObjectsPair ( FirstType firstObject, SecondType secondObject );
	explicit ObjectsPair ( std::pair < FirstType, SecondType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsPair & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Pair";

		return xmlTagName;
	}

	static ObjectsPair parse ( std::deque < sax::Token >::iterator & input );
	static std::pair < FirstType, SecondType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void compose ( std::deque < sax::Token > & out, const std::pair < FirstType, SecondType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	static std::pair < alib::Object, alib::Object > normalizeRaw ( std::pair < FirstType, SecondType > && source ) {
		return std::make_pair ( alib::Object ( alib::AnyObject < FirstType > ( std::move ( source.first ) ) ), alib::Object ( alib::AnyObject < SecondType > ( std::move ( source.second ) ) ) );
	}

	typedef ObjectsPair < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsPair < alib::Object, alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(std::pair<FirstType, SecondType> raw) : std::pair<FirstType, SecondType>(std::move(raw)) {

}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(FirstType firstObject, SecondType secondObject) : std::pair<FirstType, SecondType>(std::move(firstObject), std::move(secondObject)) {

}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::clone() const {
	return new ObjectsPair(*this);
}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::plunder() && {
	return new ObjectsPair(std::move(*this));
}

template < class FirstType, class SecondType >
int ObjectsPair < FirstType, SecondType >::compare(const ObjectsPair& other) const {
	static ext::compare<std::pair<FirstType, SecondType>> comp;
	return comp ( static_cast < const std::pair < FirstType, SecondType > & > ( * this ), static_cast < const std::pair < FirstType, SecondType > & > ( other ) );
}

template < class FirstType, class SecondType >
void ObjectsPair < FirstType, SecondType >::operator>>(std::ostream& os) const {
	os << "(ObjectsPair " << static_cast < const std::pair < FirstType, SecondType > & > ( * this ) << ")";
}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType > ObjectsPair < FirstType, SecondType >::parse(std::deque<sax::Token>::iterator& input) {
	return ObjectsPair < FirstType, SecondType > ( parseRaw ( input ) );
}

template < class FirstType, class SecondType >
void ObjectsPair < FirstType, SecondType >::compose(std::deque<sax::Token>& out) const {
	compose(out, *this);
}

template < class FirstType, class SecondType >
std::pair < FirstType, SecondType > ObjectsPair < FirstType, SecondType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsPair::getXmlTagName() );

	FirstType first = alib::xmlApi < FirstType >::parse ( input );
	SecondType second = alib::xmlApi < SecondType >::parse ( input );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsPair::getXmlTagName() );

	return std::make_pair ( std::move ( first ), std::move ( second ) );
}

template < class FirstType, class SecondType >
void ObjectsPair < FirstType, SecondType >::compose ( std::deque < sax::Token > & out, const std::pair < FirstType, SecondType > & input ) {
	out.emplace_back ( ObjectsPair::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	alib::xmlApi < typename std::decay < FirstType >::type >::compose ( out, input.first );
	alib::xmlApi < SecondType >::compose ( out, input.second );

	out.emplace_back ( ObjectsPair::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class FirstType, class SecondType >
alib::ObjectBase* ObjectsPair < FirstType, SecondType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < std::pair < T, R > > {
	static std::pair < T, R > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( std::deque < sax::Token > & output, const std::pair < T, R > & data );
};

template < typename T, typename R >
std::pair < T, R > xmlApi < std::pair < T, R > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsPair < T, R >::parseRaw ( input );
}

template < typename T, typename R >
bool xmlApi < std::pair < T, R > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName() );
}

template < typename T, typename R >
std::string xmlApi < std::pair < T, R > >::xmlTagName ( ) {
	return container::ObjectsPair < >::getXmlTagName();
}

template < typename T, typename R >
void xmlApi < std::pair < T, R > >::compose ( std::deque < sax::Token > & output, const std::pair < T, R > & input ) {
	return container::ObjectsPair < T, R >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_PAIR_H_ */
