/*
 * Components.hpp
 *
 *  Created on: Mar 16, 2016
 *      Author: Jan Travnicek
 */

#ifndef COMPONENTS_HPP_
#define COMPONENTS_HPP_

#include <set>
#include <algorithm>
#include <utility>
#include <tuple>
#include <typeinfo>
#include "../exception/CommonException.h"

namespace alib {

template < class Derived, class DataType, class SetType >
class ComponentConstraint {
public:
	/**
	 * Checks whether a concrete element is used in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is used
	 *         false if element is not used
	 */
	static bool used ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete element is available in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	static bool available ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete element is valid in context of the datatype where the set is used
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	static void valid ( const Derived & object, const DataType & symbol );
};

/**
 * Represents a set of elements.
 * @param Derived class representing datatype using this set.
 * @param DataType underlying type of symbols in the set.
 * @param SetType arbitrary type used to distinguish different sets.
 */
template < class Derived, class DataType, class SetType >
class Component {
	/**
	 * The set.
	 */
	std::set < DataType > m_data;

	/**
	 * Checks whether symbol can be added to the set. Calls valid and available functions.
	 * @throws CommonException if symbol cannot be added.
	 */
	void checkAdd ( const DataType & symbol ) {
		ComponentConstraint < Derived, DataType, SetType >::valid ( static_cast < const Derived & > ( * this ), symbol );

		if ( !ComponentConstraint < Derived, DataType, SetType >::available ( static_cast < const Derived & > ( * this ), symbol ) ) {
			std::string elementTypeName ( ext::to_string < SetType * > ( ) );
			elementTypeName.back ( ) = ' ';
			throw::exception::CommonException ( elementTypeName + "element " + ext::to_string ( symbol ) + " is not available." );
		}
	}

	/**
	 * Checks whether symbol can be removed from the set. Calls used function.
	 * @throws CommonException if symbol cannot be removed.
	 */
	void checkRemove ( const DataType & symbol ) {
		if ( ComponentConstraint < Derived, DataType, SetType >::used ( static_cast < const Derived & > ( * this ), symbol ) ) {
			std::string elementTypeName ( ext::to_string < SetType * > ( ) );
			elementTypeName.back ( ) = ' ';
			throw::exception::CommonException ( elementTypeName + "element " + ext::to_string ( symbol ) + " is used." );
		}
	}

protected:
	/** Checks the state of the set.
	 */
	void checkState ( ) {
		for ( const DataType & symbol : m_data )
			checkAdd ( symbol );
	}

public:
	/**
	 * Constructs a set containing given elements.
	 * @throw CommonException if elements are not available in context of datatype where the set is used
	 */
	Component ( std::set < DataType > symbols ) : m_data ( std::move ( symbols ) ) {
	}

	/**
	 * Adds an elements to the set.
	 * @param element to add to the set
	 * @throw CommonException if element is not available in context of datatype where the set is used
	 * @return true if element was indeed added
	 *         false if element was present in the set
	 */
	bool add ( DataType symbol ) {
		checkAdd ( symbol );
		return m_data.insert ( std::move ( symbol ) ).second;
	}

	/**
	 * Adds a set of elements to the set.
	 * @param elements to add to the set
	 * @throw CommonException if one of the elements is not available in context of datatype where the set is used
	 */
	void add ( std::set < DataType > symbols ) {
		for ( DataType symbol : std::make_moveable_set ( symbols ) )
			add ( std::move ( symbol ) );
	}

	/**
	 * Changes the set.
	 * @param elements by which to replace those currently in the set
	 * @throw CommonException if one of the removed elements is used in context of datatype where the set is used
	 *        CommonException if one of the added elements is not available in context of datatype where the set is used
	 */
	void set ( std::set < DataType > symbols ) {
		std::set < DataType > removed;
		std::set_difference ( m_data.begin ( ), m_data.end ( ), symbols.begin ( ), symbols.end ( ), std::inserter ( removed, removed.end ( ) ) );

		for ( const DataType & symbol : removed )
			checkRemove ( symbol );

		for ( const DataType & symbol : symbols )
			checkAdd ( symbol );

		m_data = std::move ( symbols );
	}

	/**
	 * @return the set.
	 */
	std::set < DataType > & get ( ) {
		return m_data;
	}

	/**
	 * @return the set.
	 */
	const std::set < DataType > & get ( ) const {
		return m_data;
	}

	/**
	 * Removes an element from the set if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 * @return true if element was indeed removed
	 *         false if element was not present in the set
	 */
	bool remove ( const DataType & symbol ) {
		checkRemove ( symbol );
		return m_data.erase ( symbol );
	}

	/**
	 * Removes a set of elements from alphabet if not used.
	 * @throw CommonException if element is used in context of datatype where the set is used
	 */
	void remove ( const std::set < DataType > & symbols ) {
		for ( const DataType & symbol : symbols )
			remove ( symbol );
	}

	/**
	 * Component emptiness checker.
	 * @return true if set is an empty
	 */
	bool empty ( ) const {
		return m_data.empty ( );
	}

};

template < class Derived, class DataType, class ElementType >
class ElementConstraint {
public:
	/**
	 * Checks whether a concrete symbol is available in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @return true if element is available
	 *         false if element is not available
	 */
	static bool available ( const Derived & object, const DataType & symbol );

	/**
	 * Checks whether a concrete symbol is valid in context of the datatype instance using this class
	 *
	 * To be implemented by all template instantiations explicitly
	 *
	 * @param element to check
	 * @throw CommonException if the element in any way invalid
	 */
	static void valid ( const Derived & object, const DataType & symbol );
};

/**
 * Represents an notable element.
 * @param Derived class representing datatype using this notable element.
 * @param DataType underlying type of element.
 * @param ElementType arbitrary type used to distinguish different notable elements.
 */
template < class Derived, class DataType, class ElementType >
class Element {
	/**
	 * The element.
	 */
	DataType m_data;

	/**
	 * Checks whether element can be set. Calls valid and available functions.
	 * @throws CommonException if element cannot be added.
	 */
	void checkSet ( const DataType & symbol ) {
		ElementConstraint < Derived, DataType, ElementType >::valid ( static_cast < const Derived & > ( * this ), symbol );

		if ( !ElementConstraint < Derived, DataType, ElementType >::available ( static_cast < const Derived & > ( * this ), symbol ) ) {
			std::string elementTypeName ( ext::to_string < ElementType * > ( ) );
			elementTypeName.back ( ) = ' ';
			throw::exception::CommonException ( elementTypeName + ext::to_string ( symbol ) + " is not available." );
		}
	}

protected:
	/** Checks the state of the element.
	 */
	void checkState ( ) {
		checkSet ( m_data );
	}

public:
	/**
	 * Constructs a notable element.
	 * @throw CommonException if element is not available in context of datatype where the class is used
	 */
	Element ( DataType symbol ) : m_data ( std::move ( symbol ) ) {
	}

	/**
	 * Changes the notable element.
	 * @param new value of the element
	 * @return bool true if the element was set
	 *              false if the element was the same as already present
	 * @throw CommonException if the new element is not available in context of datatype where the class is used
	 */
	bool set ( DataType symbol ) {
		checkSet ( symbol );

		if ( m_data == symbol ) return false;

		m_data = std::move ( symbol );
		return true;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	DataType & get ( ) {
		return m_data;
	}

	/**
	 * Returns the current notable element of ElementType.
	 * @return the notable element
	 */
	const DataType & get ( ) const {
		return m_data;
	}

};

/**
 * Auxiliary base handling all sets of elements from components
 */
template < class Derived, class DataType, class ... SetTypes >
struct ComponentAux;

/**
 * Specialisation for tuple.
 */
template < class Derived, class DataType, class ... SetTypes >
struct ComponentAux < Derived, DataType, std::tuple < SetTypes ... > > : public Component < Derived, DataType, SetTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealSetTypes, size_t ... Indexes >
	ComponentAux ( std::tuple < RealSetTypes ... > params, std::index_sequence < Indexes ... > ) : Component < Derived, DataType, SetTypes > ( std::move ( std::get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	const Component < Derived, DataType, SetType > & accessComponent ( ) const {
		return static_cast < const Component < Derived, DataType, SetType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param SetType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class SetType, typename std::enable_if < std::is_in < SetType, SetTypes ... >::value >::type * = nullptr >
	Component < Derived, DataType, SetType > & accessComponent ( ) {
		return static_cast < Component < Derived, DataType, SetType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Component < Derived, DataType, SetTypes >::checkState ( ), 0 ) ... };
	}

};

/**
 * Auxiliary base handling all notable elements from components
 */
template < class Derived, class DataType, class ... ElementTypes >
struct ElementAux;

/**
 * Specialisation for tuple.
 */
template < class Derived, class DataType, class ... ElementTypes >
struct ElementAux < Derived, DataType, std::tuple < ElementTypes ... > > : public Element < Derived, DataType, ElementTypes > ... {

	/**
	 * Constructor
	 */
	template < class ... RealElementTypes, size_t ... Indexes >
	ElementAux ( std::tuple < RealElementTypes ... > params, std::index_sequence < Indexes ... > ) : Element < Derived, DataType, ElementTypes > ( std::move ( std::get < Indexes > ( params ) ) ) ... {
		( void ) params; // No-op
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	const Element < Derived, DataType, ElementType > & accessElement ( ) const {
		return static_cast < const Element < Derived, DataType, ElementType > & > ( * this );
	}

	/**
	 * Allows access to sub-component using its type.
	 * @param ElementType alphabet type used to distinguish different sub alphabets
	 * @return sub-component
	 */
	template < class ElementType, typename std::enable_if < std::is_in < ElementType, ElementTypes ... >::value >::type * = nullptr >
	Element < Derived, DataType, ElementType > & accessElement ( ) {
		return static_cast < Element < Derived, DataType, ElementType > & > ( * this );
	}

protected:
	/**
	 * postponed checker function
	 */
	void checkState ( ) {
		( void ) std::initializer_list < int > { ( Element < Derived, DataType, ElementTypes >::checkState ( ), 0 ) ... };
	}

};

template < class ... Types >
class Components;

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived >
class Components < Derived > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	Components ( ) {
	}

	template < class T, typename std::enable_if < ! std::is_same < T, T >::value >::type * = nullptr >
	void accessComponent ( );
	template < class T, typename std::enable_if < ! std::is_same < T, T >::value >::type * = nullptr >
	void accessElement ( );
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class SetTypesPack, class ElementTypesPack, class ... NextGroup >
class Components < Derived, DataType, SetTypesPack, ElementTypesPack, NextGroup ... > : public ComponentAux < Derived, DataType, SetTypesPack >, public ElementAux < Derived, DataType, ElementTypesPack >, public Components < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class RealElementTypes, class ... NextRealTypes >
	Components ( RealSetTypes params1, RealElementTypes params2, NextRealTypes ... nextParams ) : ComponentAux < Derived, DataType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), ElementAux < Derived, DataType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ), Components < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
		ComponentAux < Derived, DataType, SetTypesPack >::checkState ( );

		ElementAux < Derived, DataType, ElementTypesPack >::checkState ( );
	}

	using ComponentAux < Derived, DataType, SetTypesPack >::accessComponent;
	using ElementAux < Derived, DataType, ElementTypesPack >::accessElement;

	using Components < Derived, NextGroup ... >::accessComponent;
	using Components < Derived, NextGroup ... >::accessElement;
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class ElementTypesPack, class ... NextGroup >
class Components < Derived, DataType, std::tuple < >, ElementTypesPack, NextGroup ... > : public ElementAux < Derived, DataType, ElementTypesPack >, public Components < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealElementTypes, class ... NextRealTypes >
	Components ( const std::tuple < > & , RealElementTypes params2, NextRealTypes ... nextParams ) : ElementAux < Derived, DataType, ElementTypesPack > ( std::move ( params2 ), std::make_index_sequence < std::tuple_size < ElementTypesPack >::value > { } ), Components < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
		ElementAux < Derived, DataType, ElementTypesPack >::checkState ( );
	}

	using ElementAux < Derived, DataType, ElementTypesPack >::accessElement;

	using Components < Derived, NextGroup ... >::accessComponent;
	using Components < Derived, NextGroup ... >::accessElement;
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class SetTypesPack, class ... NextGroup >
class Components < Derived, DataType, SetTypesPack, std::tuple < >, NextGroup ... > : public ComponentAux < Derived, DataType, SetTypesPack >, public Components < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class RealSetTypes, class ... NextRealTypes >
	Components ( RealSetTypes params1, const std::tuple < > & , NextRealTypes ... nextParams ) : ComponentAux < Derived, DataType, SetTypesPack > ( std::move ( params1 ), std::make_index_sequence < std::tuple_size < SetTypesPack >::value > { } ), Components < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
		ComponentAux < Derived, DataType, SetTypesPack >::checkState ( );
	}

	using ComponentAux < Derived, DataType, SetTypesPack >::accessComponent;

	using Components < Derived, NextGroup ... >::accessComponent;
	using Components < Derived, NextGroup ... >::accessElement;
};

/**
 * Auxiliary class allowing simple access to the alphabets.
 */
template < class Derived, class DataType, class ... NextGroup >
class Components < Derived, DataType, std::tuple < >, std::tuple < >, NextGroup ... > : public Components < Derived, NextGroup ... > {
public:
	/**
	 * Construct an alphabet pack from two alphabets.
	 */
	template < class ... NextRealTypes >
	Components ( const std::tuple < > & , const std::tuple < > & , NextRealTypes ... nextParams ) : Components < Derived, NextGroup ... > ( std::move ( nextParams ) ... ) {
	}

	using Components < Derived, NextGroup ... >::accessComponent;
	using Components < Derived, NextGroup ... >::accessElement;
};

} /* namespace alib */

#endif /* COMPONENTS2_HPP_ */
