/*
 * multipleDispatch.hpp
 *
 *  Created on: Apr 05, 2014
 *	  Author: Jan Travnicek
 */

#ifndef MULTIPLE_DISPATCH_H_
#define MULTIPLE_DISPATCH_H_

#include <stdexcept>
#include <functional>
#include <map>
#include <iostream>
#include <type_traits>
#include <sstream>
#include <tuple>
#include <memory>

#include <core/castApi.hpp>
#include <exception/CommonException.h>

namespace alib {

template < class Algorithm, class ReturnType, class ... Params >
class MultipleDispatch;

template < class Algorithm, class ReturnType, class ... FrontStaticParamTypes, class ... DispatchedParameterTypes, class ... BackStaticParamTypes >
class MultipleDispatch < Algorithm, ReturnType, std::tuple < FrontStaticParamTypes ... >, std::tuple < DispatchedParameterTypes ... >, std::tuple < BackStaticParamTypes ... > > {
public:
	template < class RealReturnType, class ... RealParameterTypeBases >
	using overload = std::function < RealReturnType ( FrontStaticParamTypes ..., typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type && ..., BackStaticParamTypes ... ) >;

	template < class RealReturnType, class ... RealParameterTypeBases >
	using rawOverload = RealReturnType ( * ) ( FrontStaticParamTypes ..., typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type && ..., BackStaticParamTypes ... );

	typedef std::tuple < FrontStaticParamTypes ... > frontParams;
	typedef std::tuple < DispatchedParameterTypes ... > dispatchedParams;
	typedef std::tuple < BackStaticParamTypes ... > backParams;
	typedef ReturnType returnType;

private:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( FrontStaticParamTypes ..., DispatchedParameterTypes && ..., BackStaticParamTypes ... ) = 0;
	};

	template < class RealReturnType, class ... RealParameterTypeBases >
	class RegistratorWrapper : public RegistratorWrapperBase {
		overload < RealReturnType, RealParameterTypeBases ... > m_callback;

	public:
		ReturnType eval ( FrontStaticParamTypes ... front, DispatchedParameterTypes && ... dispatched, BackStaticParamTypes ... back ) {
			return ReturnType ( m_callback ( front ..., std::forward < typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type > ( static_cast < typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type && > ( dispatched ) ) ..., back ... ) );
		}

		RegistratorWrapper ( overload < RealReturnType, RealParameterTypeBases ... > callback ) : m_callback ( callback ) {
		}

	};

	std::map < typename std::TupleBuilder < std::type_index, sizeof ... ( DispatchedParameterTypes ) >::type, std::unique_ptr < RegistratorWrapperBase > > registeredFunctions;

	static MultipleDispatch < Algorithm, ReturnType, std::tuple < FrontStaticParamTypes ... >, std::tuple < DispatchedParameterTypes ... >, std::tuple < BackStaticParamTypes ... > > & getInstance ( ) {
		static MultipleDispatch < Algorithm, ReturnType, std::tuple < FrontStaticParamTypes ... >, std::tuple < DispatchedParameterTypes ... >, std::tuple < BackStaticParamTypes ... > > res;

		return res;
	}

public:
	template < class RealReturnType, class ... RealParameterTypeBases >
//	static void registerOverload ( overload < RealReturnType, RealParameterTypeBases ... > callback ) { causes clang 4.0.0 to crash...
//	static void registerOverload ( rawOverload < RealReturnType, RealParameterTypeBases ... > callback ) { causes clang 4.0.0 to crash...
	static void registerOverload ( RealReturnType ( * callback ) ( FrontStaticParamTypes ..., typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type && ..., BackStaticParamTypes ... ) ) {
		auto key = std::make_tuple ( std::type_index ( typeid ( typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type ) ) ... );
		std::unique_ptr < RegistratorWrapperBase > value ( new RegistratorWrapper < RealReturnType, RealParameterTypeBases ... > ( callback ) );

		bool res = getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::move ( key ), std::move ( value ) ) ).second;

		if ( ! res ) {
			std::stringstream ss;
			( void ) std::initializer_list < int > { ( ss << ext::to_string < typename std::match_cv_ref < DispatchedParameterTypes, RealParameterTypeBases >::type > ( ), 0 ) ... };

			std::string classType = ext::to_string < Algorithm > ( );

			throw::exception::CommonException ( "Callback for " + ss.str ( ) + " already registered on " + classType + "." );
		}
	}

	static ReturnType dispatch ( FrontStaticParamTypes ... front, DispatchedParameterTypes && ... dispatched, BackStaticParamTypes ... back ) {
		auto callback = getInstance ( ).registeredFunctions.find ( std::make_tuple ( std::type_index ( typeid ( dispatched ) ) ... ) );

		if ( callback == getInstance ( ).registeredFunctions.end ( ) ) {
			std::stringstream ss;
			( void ) std::initializer_list < int > { ( ss << ext::to_string ( typeid ( dispatched ) ), 0 ) ... };

			std::string classType = ext::to_string < Algorithm > ( );

			throw::exception::CommonException ( "Callback for " + ss.str ( ) + " not registered on " + classType + "." );
		}

		return callback->second->eval ( front ..., std::forward < DispatchedParameterTypes > ( dispatched ) ..., back ... );
	}

};

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class Algorithm, class ReturnType, class FirstParameterType, class ... StaticParamTypes >
using SingleDispatch = MultipleDispatch < Algorithm, ReturnType, std::tuple < >, std::tuple < FirstParameterType >, std::tuple < StaticParamTypes ... > >;

template < class Algorithm, class ReturnType, class StaticParamType, class FirstParameterType >
using SingleDispatchFirstStaticParam = MultipleDispatch < Algorithm, ReturnType, std::tuple < StaticParamType >, std::tuple < FirstParameterType >, std::tuple < > >;

template < class Algorithm, class ReturnType, class FirstParameterType, class SecondParameterType, class ... StaticParamTypes >
using DoubleDispatch = MultipleDispatch < Algorithm, ReturnType, std::tuple < >, std::tuple < FirstParameterType, SecondParameterType >, std::tuple < StaticParamTypes ... > >;

template < class Algorithm, class ReturnType, class StaticParamType, class FirstParameterType, class SecondParameterType >
using DoubleDispatchFirstStaticParam = MultipleDispatch < Algorithm, ReturnType, std::tuple < StaticParamType >, std::tuple < FirstParameterType, SecondParameterType >, std::tuple < > >;

// --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

template < class Algorithm, class ReturnType, class ParametersType >
class PromotingDoubleDispatch {
public:
	template < class RealReturnType, class RealParametersTypeBase >
	using overload = std::function < RealReturnType ( typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type &&, typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type && ) >;

	template < class RealReturnType, class RealParametersTypeBase >
	using rawOverload = RealReturnType ( * ) ( typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type &&, typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type && );

	typedef std::tuple < > frontParams;
	typedef std::tuple < ParametersType, ParametersType > dispatchedParams;
	typedef std::tuple < > backParams;
	typedef ReturnType returnType;

private:
	class RegistratorWrapperBase {
	public:
		virtual ReturnType eval ( bool firstAttempt, ParametersType &&, ParametersType && ) = 0;
		virtual bool available ( bool firstAttempt, std::type_index first, std::type_index second ) = 0;
	};

	template < class RealReturnType, class RealParametersTypeBase >
	class RegistratorWrapper : public RegistratorWrapperBase {
		typedef typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type RealParametersType;

		std::function < RealReturnType ( RealParametersType &&, RealParametersType && ) > m_callback;

	public:
		ReturnType eval ( bool firstAttempt, ParametersType && first, ParametersType && second ) {
			if ( std::type_index ( typeid ( first ) ) == std::type_index ( typeid ( second ) ) )
				return ReturnType ( m_callback ( std::forward < RealParametersType > ( static_cast < RealParametersType && > ( first ) ), std::forward < RealParametersType > ( static_cast < RealParametersType && > ( second ) ) ) );

			if ( firstAttempt ) {
				alib::Object casted = alib::castApi::getCastPool ( std::type_index ( typeid ( first ) ) ).cast ( std::forward < ParametersType > ( second ) );
				return ReturnType ( m_callback ( std::forward < RealParametersType > ( static_cast < RealParametersType && > ( first ) ), std::forward < RealParametersType > ( static_cast < RealParametersType && > ( casted.getData ( ) ) ) ) );
			} else {
				alib::Object casted = alib::castApi::getCastPool ( std::type_index ( typeid ( second ) ) ).cast ( std::forward < ParametersType > ( first ) );
				return ReturnType ( m_callback ( std::forward < RealParametersType > ( static_cast < RealParametersType && > ( casted.getData ( ) ) ), std::forward < RealParametersType > ( static_cast < RealParametersType && > ( second ) ) ) );
			}
		}

		bool available ( bool firstAttempt, std::type_index first, std::type_index second ) {
			if ( first == second ) return true;

			if ( firstAttempt )
				return alib::castApi::castAvailable ( first, second );
			else
				return alib::castApi::castAvailable ( second, first );
		}

		RegistratorWrapper ( RealReturnType ( * callback ) ( RealParametersType &&, RealParametersType && ) ) : m_callback ( callback ) {
		}

	};

	std::map < std::type_index, std::unique_ptr < RegistratorWrapperBase > > registeredFunctions;

	static PromotingDoubleDispatch < Algorithm, ReturnType, ParametersType > & getInstance ( ) {
		static PromotingDoubleDispatch < Algorithm, ReturnType, ParametersType > res;

		return res;
	}

public:
	template < class RealReturnType, class RealParametersTypeBase >
	static void registerOverload ( RealReturnType ( * callback ) ( typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type &&, typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type && ) ) {
		std::unique_ptr < RegistratorWrapperBase > value ( new RegistratorWrapper < RealReturnType, RealParametersTypeBase > ( callback ) );

		bool res = getInstance ( ).registeredFunctions.insert ( std::make_pair ( std::type_index ( typeid ( typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type ) ), std::move ( value ) ) ).second;
		if ( ! res ) {
			std::string paramsType = ext::to_string < typename std::match_cv_ref < ParametersType, RealParametersTypeBase >::type > ( );

			std::string classType = ext::to_string < Algorithm > ( );

			throw::exception::CommonException ( "Callback for " + paramsType + " already registered on " + classType + "." );
		}
	}

	static ReturnType dispatch ( ParametersType && first, ParametersType && second ) {
		typename std::map < std::type_index, std::unique_ptr < RegistratorWrapperBase > >::iterator callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( first ) ) );

		if ( ( callback != getInstance ( ).registeredFunctions.end ( ) ) && callback->second->available ( true, std::type_index ( typeid ( first ) ), std::type_index ( typeid ( second ) ) ) )
			return callback->second->eval ( true, std::forward < ParametersType > ( first ), std::forward < ParametersType > ( second ) );

		callback = getInstance ( ).registeredFunctions.find ( std::type_index ( typeid ( second ) ) );

		if ( ( callback != getInstance ( ).registeredFunctions.end ( ) ) && callback->second->available ( false, std::type_index ( typeid ( first ) ), std::type_index ( typeid ( second ) ) ) )
			return callback->second->eval ( false, std::forward < ParametersType > ( first ), std::forward < ParametersType > ( second ) );

		std::string firstType = ext::to_string ( typeid ( first ) );

		std::string secondType = ext::to_string ( typeid ( second ) );

		std::string classType = ext::to_string < Algorithm > ( );

		throw::exception::CommonException ( "Callback for (" + firstType + ", " + secondType + ") (promoting) not registered on " + classType + "." );
	}

};

} /* namespace alib */

#endif /* MULTIPLE_DISPATCH_H_ */
