/*
 * XmlApi.hpp
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef XML_API_HPP_
#define XML_API_HPP_

#include <functional>
#include <deque>
#include <vector>
#include <map>
#include <string>
#include <memory>
#include "sax/Token.h"
#include "sax/FromXMLParserHelper.h"

#include "object/ObjectBase.h"
#include "base/WrapperBase.hpp"
#include "exception/CommonException.h"

#include <global/GlobalData.h>

namespace alib {

class xmlApiInputContext : public std::deque < sax::Token >::iterator {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < int, WrapperBaseBase * > > idToInstanceContexts;
	int idMaxContext;

public:
	xmlApiInputContext ( std::deque < sax::Token >::iterator iter ) : std::deque < sax::Token >::iterator ( iter ), idMaxContext ( 0 ) {
	}

	~xmlApiInputContext ( ) {
		for ( const std::pair < const std::string, std::map < int, WrapperBaseBase * > > context : idToInstanceContexts )
			for ( const std::pair < const int, WrapperBaseBase * > entry : context.second )
				delete entry.second;
	}

	std::map < int, WrapperBaseBase * > & idToInstance ( const std::string & name ) {
		return idToInstanceContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

	const std::map < std::string, std::map < int, WrapperBaseBase * > > & dump ( ) const {
		return idToInstanceContexts;
	}

};

class xmlApiOutputContext : public std::deque < sax::Token > {
	template < typename T, typename Enable >
	friend struct xmlApi;

	std::map < std::string, std::map < const CommonBaseBase *, int > > instanceToIdContexts;
	int idMaxContext;

public:
	xmlApiOutputContext ( ) : idMaxContext ( 0 ) {
	}

	std::map < const CommonBaseBase *, int > & instanceToId ( const std::string & name ) {
		return instanceToIdContexts[name];
	}

	int & idMax ( ) {
		return idMaxContext;
	}

};

template < typename T, typename Enable = void >
struct xmlApi { };

template < typename Type >
struct xmlApi < Type, typename std::enable_if < std::is_base_of < ObjectBase, Type >::value >::type > {
	static Type parse ( std::deque < sax::Token >::iterator & input ) {
		return Type::parse ( input );
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName() );
	}

	static std::string xmlTagName ( ) {
		return Type::getXmlTagName();
	}

	static void compose ( std::deque < sax::Token > & output, const Type & data ) {
		data.compose ( output );
	}

};

template < typename Group >
struct xmlApi < Group, typename std::enable_if < std::is_base_of < WrapperBaseBase, Group >::value >::type > {
private:
	class GroupParser {
	public:
		virtual Group parse ( std::deque < sax::Token >::iterator & input ) = 0;
	};

	static std::map < std::string, std::unique_ptr < GroupParser > > & parseFunctions ( ) {
		static std::map < std::string, std::unique_ptr < GroupParser > > res;

		return res;
	}

	template < class Type >
	class ParserRegister : public GroupParser {
		std::function < Type ( std::deque < sax::Token >::iterator & ) > parseFunction;

	public:
		ParserRegister( ) : parseFunction ( Type::parse ) {
		}

		virtual Group parse ( std::deque < sax::Token >::iterator & input ) {
			return Group ( parseFunction ( input ) );
		}

	};

public:
	template < class Type >
	static void registerType ( ) {
		bool res = parseFunctions ( ).insert ( std::make_pair ( xmlApi < Type >::xmlTagName(), std::unique_ptr < GroupParser > ( new ParserRegister < Type > ( ) ) ) ).second;
		if ( ! res ) {
			std::string groupName = ext::to_string < Group > ( );
			std::string typeName = ext::to_string < Type > ( );

			throw::exception::CommonException ( "Parse callback of " + typeName + " already registered in group " + groupName + "." );
		}
	}

	static Group parse ( std::deque < sax::Token >::iterator & data ) {
		xmlApiInputContext & input = ( xmlApiInputContext & ) data;
		sax::FromXMLParserHelper::skipAttributes ( input, sax::Token::TokenType::START_ELEMENT );

		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );
		if ( /* common::GlobalData::optimizeXml && */ tagName == "Ref" ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Ref" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ATTRIBUTE, "id" );
			int id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( input, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ATTRIBUTE, "id" );
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Ref" );
			std::map < int, WrapperBaseBase * >::iterator elem = input.idToInstance ( ext::to_string < Group > ( ) ).find ( id );

			if ( elem == input.idToInstance ( ext::to_string < Group > ( ) ).end ( ) ) {
				std::cerr << input.dump ( ) << std::endl;
				throw exception::CommonException ( "XML Inconsistent ( id not found " + ext::to_string ( id  ) + " )" );
			}

			return * ( ( Group * ) elem->second );
		} else {
			typename std::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

			if ( callback == parseFunctions ( ).end ( ) ) throw exception::CommonException ( "Parse callback for " + tagName + " tag not registered." );

			/* find out if this is a base for reference */
			bool ref = sax::FromXMLParserHelper::isToken ( input + 1, sax::Token::TokenType::START_ATTRIBUTE, "ref" );
			int id = 0;
			if ( ref ) {
				std::deque < sax::Token >::iterator tmp = input + 2;
				id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( tmp, sax::Token::TokenType::CHARACTER ) );
			}

			Group res ( callback->second->parse ( input ) );

			/* if object is a base of reference, register it */
			if ( ref )
				input.idToInstance ( ext::to_string < Group > ( ) ).insert ( std::make_pair ( id, ( WrapperBaseBase * ) new Group ( res ) ) );

			return res;
		}
	}

	static bool first ( const std::deque < sax::Token >::const_iterator & input ) {
		const std::string & tagName = sax::FromXMLParserHelper::getTokenData ( input, sax::Token::TokenType::START_ELEMENT );

		if ( tagName == "Ref" ) return true;

		typename std::map < std::string, std::unique_ptr < GroupParser > >::iterator callback = parseFunctions ( ).find ( tagName );

		if ( callback == parseFunctions ( ).end ( ) ) return false;

		return true;
	}

	static std::string xmlTagName ( ) {
		std::string target = ext::to_string < Group > ( );

		throw ::exception::CommonException ( "Type " + target + " does not have xmlTagName." );
	}

	static void compose ( std::deque < sax::Token > & output, const Group & data ) {
		xmlApiOutputContext & context = ( xmlApiOutputContext & ) output;

		typename std::map < const CommonBaseBase *, int >::iterator elem = context.instanceToId ( ext::to_string < Group > ( ) ).find ( static_cast < const CommonBaseBase * > ( & data.getData ( ) ) );

		if ( common::GlobalData::optimizeXml && elem != context.instanceToId ( ext::to_string < Group > ( ) ).end ( ) ) {
			output.emplace_back ( "Ref", sax::Token::TokenType::START_ELEMENT );
			output.emplace_back ( "id", sax::Token::TokenType::START_ATTRIBUTE );
			output.emplace_back ( ext::to_string ( elem->second ), sax::Token::TokenType::CHARACTER );
			output.emplace_back ( "id", sax::Token::TokenType::END_ATTRIBUTE );
			output.emplace_back ( "Ref", sax::Token::TokenType::END_ELEMENT );
		} else {
			/* get current position in token stream */
			size_t pos = output.size();
			int id = 0;
			if ( common::GlobalData::optimizeXml ) {
				id = context.idMax ( )++;
			}

			/* encode referenced object */
			data.getData ( ).compose ( output );

			/* emplace referenced attribute */
			if ( common::GlobalData::optimizeXml ) {
				output.emplace ( output.begin ( ) + pos + 1, "ref", sax::Token::TokenType::START_ATTRIBUTE );
				output.emplace ( output.begin ( ) + pos + 2, ext::to_string ( id ), sax::Token::TokenType::CHARACTER );
				output.emplace ( output.begin ( ) + pos + 3, "ref", sax::Token::TokenType::END_ATTRIBUTE );

				context.instanceToId ( ext::to_string < Group > ( ) ).insert ( std::make_pair ( static_cast< const CommonBaseBase * > ( & data.getData ( ) ), id ) );
			}
		}
	}

};

} /* namespace alib */

#endif /* XML_API_HPP_ */
