/*
 * DFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef DFA_H_
#define DFA_H_

#include <map>
#include <ostream>
#include <sstream>
#include <range>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

template<class SymbolType, class StateType >
class DFA final : public AutomatonBase, public alib::Components < DFA < SymbolType, StateType >, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < StateType, SymbolType >, StateType > transitions;

public:
	explicit DFA ( StateType initialState );
	explicit DFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, StateType initialState, std::set < StateType > finalStates );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this-> template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this-> template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this-> template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this-> template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this-> template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this-> template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this-> template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this-> template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this-> template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this-> template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this-> template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this-> template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this-> template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this-> template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this-> template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType current, SymbolType input, const StateType next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < StateType, SymbolType >, StateType > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::range < typename std::map < std::pair < StateType, SymbolType >, StateType >::const_iterator > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < StateType, SymbolType >, StateType > getTransitionsToState ( const StateType & from ) const;

	/**
	 * Determines whether DFA is total deterministic
	 * FA is deterministic if and only if
	 * \li \c it is deterministic. (Trivial here)
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 * @return true when automaton is total deterministic, false otherwise
	 */
	bool isTotal ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DFA";

		return xmlTagName;
	}

	static DFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef DFA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( DFA < > ) == typeid ( DFA < SymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		DFA < > * res = new DFA < > ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < std::pair < StateType, SymbolType >, StateType > && transition : std::make_moveable_map ( transitions ) ) {
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			DefaultSymbolType input = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( transition.first.second ) );
			DefaultStateType to = AutomatonNormalize::normalizeState ( std::move ( transition.second ) );

			res->addTransition ( std::move ( from ), std::move ( input ), std::move ( to ) );
		}

		return res;
	}
};

template<class SymbolType, class StateType >
DFA<SymbolType, StateType>::DFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, StateType initialState, std::set < StateType > finalStates ) : alib::Components < DFA, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template<class SymbolType, class StateType >
DFA<SymbolType, StateType>::DFA ( StateType initialState ) : DFA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, initialState, std::set < StateType > { } ) {
}

template<class SymbolType, class StateType >
AutomatonBase * DFA<SymbolType, StateType>::clone ( ) const {
	return new DFA<SymbolType, StateType> ( * this );
}

template<class SymbolType, class StateType >
AutomatonBase * DFA<SymbolType, StateType>::plunder ( ) && {
	return new DFA<SymbolType, StateType> ( std::move ( * this ) );
}

template<class SymbolType, class StateType >
bool DFA<SymbolType, StateType>::addTransition ( StateType from, SymbolType input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + ext::to_string ( input ) + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist." );

	std::pair < StateType, SymbolType > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	if ( transitions.find ( key ) != transitions.end ( ) ) {
		if ( transitions.find ( key )->second == to )
			return false;
		else
			throw AutomatonException ( "Transition from this state and symbol already exists (\"" + ext::to_string ( key.first ) + "\", \"" + ext::to_string ( key.second ) + "\") -> \"" + ext::to_string ( to ) + "\"." );
	}

	transitions.insert ( std::make_pair ( std::move ( key ), std::move ( to ) ) );
	return true;
}

template<class SymbolType, class StateType >
bool DFA<SymbolType, StateType>::removeTransition ( const StateType & from, const SymbolType & input, const StateType & to ) {
	std::pair < StateType, SymbolType > key = std::make_pair ( from, input );

	if ( transitions.find ( key ) == transitions.end ( ) )
		return false;

	if ( transitions.find ( key )->second != to )
		throw AutomatonException ( "Transition (\"" + ext::to_string ( from ) + "\", \"" + ext::to_string ( input ) + "\") -> \"" + ext::to_string ( to ) + "\" doesn't exist." );

	transitions.erase ( key );
	return true;
}

template<class SymbolType, class StateType >
const std::map < std::pair < StateType, SymbolType >, StateType > & DFA<SymbolType, StateType>::getTransitions ( ) const {
	return transitions;
}

	template < class Type, class From >
	class PartialComparator {
		const Type & m_data;
		std::function < const Type & ( const From & ) > retrieveFunction;

	public:
		PartialComparator ( const Type & data, const Type & ( * function ) ( const From & ) ) : m_data ( data ), retrieveFunction ( function ) {

		}

		friend bool operator < ( const PartialComparator & first, const Type & second ) {
			return first.m_data < first.retrieveFunction ( second );
		}

		friend bool operator < ( const Type & first, const PartialComparator & second ) {
			return second.retrieveFunction ( first ) < second.m_data;
		}
	};

template<class SymbolType, class StateType >
std::range < typename std::map < std::pair < StateType, SymbolType >, StateType >::const_iterator > DFA<SymbolType, StateType>::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );
	typename std::map < std::pair < StateType, SymbolType >, StateType >::const_iterator lower = transitions.begin ( );
	while ( lower != transitions.end ( ) && lower->first.first < from ) {
		++ lower;
	}

	typename std::map < std::pair < StateType, SymbolType >, StateType >::const_iterator upper = transitions.begin ( );
	while ( upper != transitions.end ( ) && upper->first.first <= from ) {
		++ upper;
	}

	return std::make_range ( lower, upper );
}

template<class SymbolType, class StateType >
std::map < std::pair < StateType, SymbolType >, StateType > DFA<SymbolType, StateType>::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	std::map < std::pair < StateType, SymbolType >, StateType > transitionsToState;

	for ( const std::pair < const std::pair < StateType, SymbolType >, StateType > & transition : transitions )
		if ( transition.second == to )
			transitionsToState.insert ( transition );

	return transitionsToState;
}

template<class SymbolType, class StateType >
bool DFA<SymbolType, StateType>::isTotal ( ) const {
	return transitions.size ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

template<class SymbolType, class StateType >
int DFA<SymbolType, StateType>::compare ( const DFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.transitions );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template<class SymbolType, class StateType >
void DFA<SymbolType, StateType>::operator >>( std::ostream & out ) const {
	out << "(DFA "
	    << " states = " << getStates ( )
	    << " inputAlphabet = " << getInputAlphabet ( )
	    << " initialState = " << getInitialState ( )
	    << " finalStates = " << getFinalStates ( )
	    << " transitions = " << transitions
	    << ")";
}

template<class SymbolType, class StateType >
DFA<SymbolType, StateType>::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template<class SymbolType, class StateType >
DFA<SymbolType, StateType> DFA<SymbolType, StateType>::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, DFA::getXmlTagName() );

	std::set < StateType > states = AutomatonFromXMLParser::parseStates < StateType > ( input );
	std::set < SymbolType > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < SymbolType > ( input );
	StateType initialState = AutomatonFromXMLParser::parseInitialState < StateType > ( input );
	std::set < StateType > finalStates = AutomatonFromXMLParser::parseFinalStates < StateType > ( input );

	DFA<SymbolType, StateType> automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < DFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, DFA::getXmlTagName() );
	return automaton;
}

template<class SymbolType, class StateType >
void DFA<SymbolType, StateType>::parseTransition ( std::deque < sax::Token >::iterator & input, DFA<SymbolType, StateType> & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	StateType from = AutomatonFromXMLParser::parseTransitionFrom < StateType > ( input );
	SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol < SymbolType > ( input );
	StateType to = AutomatonFromXMLParser::parseTransitionTo < StateType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputSymbol ), std::move ( to ) );
}

template<class SymbolType, class StateType >
void DFA<SymbolType, StateType>::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( DFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( DFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template<class SymbolType, class StateType >
void DFA<SymbolType, StateType>::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) ) {
		out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

		AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
		AutomatonToXMLComposer::composeTransitionInputSymbol ( out, transition.first.second );
		AutomatonToXMLComposer::composeTransitionTo ( out, transition.second );

		out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
	}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
alib::ObjectBase* DFA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace alib {

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::DFA<SymbolType, StateType>, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DFA<SymbolType, StateType> & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const std::pair < StateType, SymbolType >, StateType > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second == symbol )
				return true;

		return false;
	}

	static bool available ( const automaton::DFA<SymbolType, StateType> &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::DFA<SymbolType, StateType> &, const SymbolType & ) {
	}

};

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::DFA<SymbolType, StateType>, StateType, automaton::States > {
public:
	static bool used ( const automaton::DFA<SymbolType, StateType> & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < StateType, SymbolType >, StateType > & t : automaton.getTransitions ( ) )
			if ( ( t.first.first == state ) || ( t.second == state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::DFA<SymbolType, StateType> &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::DFA<SymbolType, StateType> &, const StateType & ) {
	}

};

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::DFA<SymbolType, StateType>, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::DFA<SymbolType, StateType> &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::DFA<SymbolType, StateType> & automaton, const StateType & state ) {
		return automaton.getStates ( ).count ( state );
	}

	static void valid ( const automaton::DFA<SymbolType, StateType> &, const StateType & ) {
	}

};

template<class SymbolType, class StateType >
class ElementConstraint< automaton::DFA<SymbolType, StateType>, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::DFA<SymbolType, StateType> & automaton, const StateType & state ) {
		return automaton.getStates ( ).count ( state );
	}

	static void valid ( const automaton::DFA<SymbolType, StateType> &, const StateType & ) {
	}

};

} /* namespace alib */

#endif /* DFA_H_ */
