/*
 * SinglePopDPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef SINGLE_POP_DPDA_H_
#define SINGLE_POP_DPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>
#include <ostream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, r) \neq \emptyset$, $\delta (q, a, s) \neq \emptyset$ then $r \neq s$.
 * if $\delta(q, a, r) \neq \emptyset$, $\delta (q, \varepsilon, s) \neq \emptyset$, then $r \neq s$.
 */
template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class SinglePopDPDA final : public AutomatonBase, public alib::Components < SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, std::tuple < InputAlphabet >, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > transitions;

public:
	explicit SinglePopDPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreSymbol, StateType initialState, PushdownStoreSymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit SinglePopDPDA ( StateType initialState, PushdownStoreSymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < PushdownStoreSymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const PushdownStoreSymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const PushdownStoreSymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < InputSymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( InputSymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const InputSymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the SinglePopDPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the SinglePopDPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, ext::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	bool addTransition ( StateType from, InputSymbolType input, PushdownStoreSymbolType pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	bool addTransition ( StateType from, PushdownStoreSymbolType pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	/**
	 * Removes the transition from the SinglePopDPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the SinglePopDPDA
	 */
	bool removeTransition ( const StateType & from, const ext::variant < EpsilonType, InputSymbolType > & input, const PushdownStoreSymbolType & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	bool removeTransition ( const StateType & from, const InputSymbolType & input, const PushdownStoreSymbolType & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	bool removeTransition ( const StateType & from, const PushdownStoreSymbolType & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	/**
	 * @return SinglePopDPDA transitions
	 */
	const std::map < std::tuple < StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > & getTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SinglePopDPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "SinglePopDPDA";

		return xmlTagName;
	}

	static SinglePopDPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, SinglePopDPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef SinglePopDPDA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( SinglePopDPDA < > ) == typeid ( SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultSymbolType > pushdownAlphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < PushdownStoreAlphabet > ( ).get ( ) ) );
		DefaultSymbolType bottomSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( this->template accessElement < InitialSymbol > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		SinglePopDPDA < > * res = new SinglePopDPDA < > ( std::move ( states ), std::move ( alphabet ), std::move ( pushdownAlphabet ), std::move ( initialState ), std::move ( bottomSymbol ), std::move ( finalStates ) );

		for ( std::pair < std::tuple < StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > && transition : std::make_moveable_map ( transitions ) ) {
			DefaultStateType to = AutomatonNormalize::normalizeState ( std::move ( transition.second.first ) );
			std::vector < DefaultSymbolType > push = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( transition.second.second ) );

			DefaultSymbolType pop = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( std::get < 2 > ( transition.first ) ) );
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( std::get < 0 > ( transition.first ) ) );
			ext::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( std::get < 1 > ( transition.first ) ) );

			res->addTransition ( std::move ( from ), std::move ( input ), std::move ( pop ), std::move ( to ), std::move ( push ) );
		}

		return res;
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::SinglePopDPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreAlphabet, StateType initialState, PushdownStoreSymbolType initialSymbol, std::set < StateType > finalStates ) : alib::Components < SinglePopDPDA, InputSymbolType, std::tuple < InputAlphabet >, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::SinglePopDPDA(StateType initialState, PushdownStoreSymbolType initialPushdownSymbol) : SinglePopDPDA ( std::set < StateType > { initialState }, std::set < InputSymbolType > { }, std::set < PushdownStoreSymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::clone() const {
	return new SinglePopDPDA(*this);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::plunder() && {
	return new SinglePopDPDA(std::move(*this));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, ext::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, StateType to, std::vector<PushdownStoreSymbolType> push) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + ext::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is < InputSymbolType >() && ! getInputAlphabet().count(input.template get < InputSymbolType >())) {
		throw AutomatonException("Input symbol \"" + ext::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + ext::to_string ( to ) + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + ext::to_string ( pop ) + "\" doesn't exist.");
	}

	for(const PushdownStoreSymbolType& pushSymbol : push) {
		if (!getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + ext::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key(std::move(from), std::move(input), std::move(pop));
	std::pair<StateType, std::vector<PushdownStoreSymbolType> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (input.template is < InputSymbolType >())
			throw AutomatonException("Transition (\"" + ext::to_string ( std::get<0>(key) ) + "\", \"" + ext::to_string ( std::get<1>(key).template get < InputSymbolType >() ) + "\", \"" + ext::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
		else
			throw AutomatonException("Transition (\"" + ext::to_string ( std::get<0>(key) ) + "\", \"" + ext::to_string ( std::get<1>(key).template get<EpsilonType>() ) + "\", \"" + ext::to_string ( std::get<2>(key) ) + "\") -> ?? already exists.");
	}

	if(input.template is<EpsilonType>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && std::get<2>(transition.first) == std::get<2>(key)) return true;
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key2(std::get<0>(key), ext::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( ), std::get<2>(key));
		if(transitions.find(key2) != transitions.end())
			throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, InputSymbolType input, PushdownStoreSymbolType pop, StateType to, std::vector<PushdownStoreSymbolType> push) {
	ext::variant < EpsilonType, InputSymbolType > inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, PushdownStoreSymbolType pop, StateType to, std::vector<PushdownStoreSymbolType> push) {
	auto inputVariant = ext::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const ext::variant < EpsilonType, InputSymbolType >& input, const PushdownStoreSymbolType& pop, const StateType& to, const std::vector<PushdownStoreSymbolType>& push) {
	std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key(from, input, pop);
	std::pair<StateType, std::vector<PushdownStoreSymbolType> > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.template is < InputSymbolType >())
			throw AutomatonException("Transition (\"" + ext::to_string ( from ) + "\", \"" + ext::to_string ( input.template get < InputSymbolType >() ) + "\") -> \"" + ext::to_string ( to ) + "\" doesn't exist.");
		else
			throw AutomatonException("Transition (\"" + ext::to_string ( from ) + "\", \"" + ext::to_string ( input.template get<EpsilonType>() ) + "\") -> \"" + ext::to_string ( to ) + "\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const InputSymbolType& input, const PushdownStoreSymbolType& pop, const StateType& to, const std::vector<PushdownStoreSymbolType>& push) {
	ext::variant < EpsilonType, InputSymbolType > inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const PushdownStoreSymbolType& pop, const StateType& to, const std::vector<PushdownStoreSymbolType>& push) {
	auto inputVariant = ext::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
const std::map<std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getTransitions() const {
	return transitions;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
int SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compare(const SinglePopDPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	static ext::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator>>(std::ostream& out) const {
	out << "(SinglePopDPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, SinglePopDPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<InputSymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<InputSymbolType>(input);
	std::set<PushdownStoreSymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<PushdownStoreSymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	PushdownStoreSymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<PushdownStoreSymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<SinglePopDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, SinglePopDPDA::getXmlTagName());
	return automaton;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, SinglePopDPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	ext::variant < EpsilonType, InputSymbolType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, InputSymbolType>(input);
	PushdownStoreSymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<PushdownStoreSymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<PushdownStoreSymbolType> push = AutomatonFromXMLParser::parseTransitionPush<PushdownStoreSymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(SinglePopDPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
alib::ObjectBase* SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace alib {

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const InputSymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is < InputSymbolType >() && symbol == std::get<1>(transition.first).template get < InputSymbolType >())
				return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions())
			if (symbol == std::get<2>(transition.first) || std::find(transition.second.second.begin(), transition.second.second.end(), symbol) != transition.second.second.end())
				return true;

		if(automaton.getInitialSymbol() == symbol)
			return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ElementConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, ext::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions())
			if (state == std::get<0>(transition.first) || transition.second.first == state)
				return true;

		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ElementConstraint< automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace alib */

#endif /* SINGLE_POP_DPDA_H_ */
