/*
 * CNF.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef CNF_H_
#define CNF_H_

#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../common/GrammarNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace grammar {

/**
 * Context free grammar in Chomsky normal form. Type 2 in Chomsky hierarchy. Produces context free languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class CNF final : public GrammarBase, public alib::Components < CNF < SymbolType >, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > rules;
	bool generatesEpsilon;

public:
	explicit CNF ( SymbolType initialSymbol );

	explicit CNF ( std::set < SymbolType > nonTerminalSymbols, std::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( SymbolType leftHandSide, ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rightHandSide );
	bool addRule ( SymbolType leftHandSide, SymbolType rightHandSide );
	bool addRule ( SymbolType leftHandSide, std::pair < SymbolType, SymbolType > rightHandSide );

	void addRules ( SymbolType leftHandSide, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > rightHandSide );

	const std::map < SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & getRules ( ) const;

	bool removeRule ( const SymbolType & leftHandSide, const ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const SymbolType & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, SymbolType > & rightHandSide );

	bool addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide );

	std::map < SymbolType, std::set < std::vector < SymbolType > > > getRawRules ( ) const;

	bool removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getNonterminalAlphabet ( ) const {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < SymbolType > & getTerminalAlphabet ( ) const {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void setGeneratesEpsilon ( bool genEps );
	bool getGeneratesEpsilon ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CNF & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "CNF";

		return xmlTagName;
	}

	static CNF parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, CNF & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef CNF < > normalized_type;

	virtual GrammarBase * normalize ( ) && {
		if ( typeid ( CNF < > ) == typeid ( CNF < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > nonterminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < NonterminalAlphabet > ( ).get ( ) ) );
		std::set < DefaultSymbolType > terminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < TerminalAlphabet > ( ).get ( ) ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( this->template accessElement < InitialSymbol > ( ).get ( ) ) );

		CNF < > * res = new CNF < > ( std::move ( nonterminals ), std::move ( terminals ), std::move ( initialSymbol ) );

		for ( std::pair < SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > && rule : std::make_moveable_map ( rules ) ) {

			std::set < ext::variant < DefaultSymbolType, std::pair < DefaultSymbolType, DefaultSymbolType > > > rhs;
			for ( ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > && target : std::make_moveable_set ( rule.second ) )
				rhs.insert ( GrammarNormalize::normalizeRHS ( std::move ( target ) ) );

			DefaultSymbolType lhs = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( rule.first ) );

			res->addRules ( std::move ( lhs ), std::move ( rhs ) );
		}

		res->setGeneratesEpsilon ( getGeneratesEpsilon ( ) );

		return res;
	}
};

template < class SymbolType >
CNF < SymbolType >::CNF ( SymbolType initialSymbol ) : CNF ( std::set < SymbolType > { initialSymbol }, std::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
CNF < SymbolType >::CNF ( std::set < SymbolType > nonterminalAlphabet, std::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : alib::Components < CNF, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ), generatesEpsilon ( false ) {
}

template < class SymbolType >
GrammarBase * CNF < SymbolType >::clone ( ) const {
	return new CNF ( * this );
}

template < class SymbolType >
GrammarBase * CNF < SymbolType >::plunder ( ) && {
	return new CNF ( std::move ( * this ) );
}

template < class SymbolType >
bool CNF < SymbolType >::addRule ( SymbolType leftHandSide, ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.template is < SymbolType > ( ) ) {
		if ( !getTerminalAlphabet ( ).count ( rightHandSide.template get < SymbolType > ( ) ) )
			throw GrammarException ( "Rule must rewrite to terminal symbol" );
	} else {
		const std::pair < SymbolType, SymbolType > rhs = rightHandSide.template get < std::pair < SymbolType, SymbolType > > ( );

		if ( !getNonterminalAlphabet ( ).count ( rhs.first ) )
			throw GrammarException ( "Symbol \"" + ext::to_string ( rhs.first ) + "\" is not a nonterminal symbol" );

		if ( !getNonterminalAlphabet ( ).count ( rhs.second ) )
			throw GrammarException ( "Symbol \"" + ext::to_string ( rhs.second ) + "\" is not a nonterminal symbol" );
	}

	return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
bool CNF < SymbolType >::addRule ( SymbolType leftHandSide, SymbolType rightHandSide ) {
	ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
bool CNF < SymbolType >::addRule ( SymbolType leftHandSide, std::pair < SymbolType, SymbolType > rightHandSide ) {
	ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
void CNF < SymbolType >::addRules ( SymbolType leftHandSide, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > & element : rightHandSide ) {
		if ( element.template is < SymbolType > ( ) ) {
			if ( !getTerminalAlphabet ( ).count ( element.template get < SymbolType > ( ) ) )
				throw GrammarException ( "Rule must rewrite to terminal symbol" );
		} else {
			const std::pair < SymbolType, SymbolType > rhs = element.template get < std::pair < SymbolType, SymbolType > > ( );

			if ( !getNonterminalAlphabet ( ).count ( rhs.first ) )
				throw GrammarException ( "Symbol \"" + ext::to_string ( rhs.first ) + "\" is not a nonterminal symbol" );

			if ( !getNonterminalAlphabet ( ).count ( rhs.second ) )
				throw GrammarException ( "Symbol \"" + ext::to_string ( rhs.second ) + "\" is not a nonterminal symbol" );
		}
	}

	rules[std::move ( leftHandSide )].insert ( std::make_moveable_set ( rightHandSide ).begin ( ), std::make_moveable_set ( rightHandSide ).end ( ) );
}

template < class SymbolType >
const std::map < SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & CNF < SymbolType >::getRules ( ) const {
	return rules;
}

template < class SymbolType >
bool CNF < SymbolType >::removeRule ( const SymbolType & leftHandSide, const ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool CNF < SymbolType >::removeRule ( const SymbolType & leftHandSide, const SymbolType & rightHandSide ) {
	ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool CNF < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, SymbolType > & rightHandSide ) {
	ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool CNF < SymbolType >::addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( true );
		return !res;
	} else if ( rightHandSide.size ( ) == 1 ) {
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide[0] ) );
	} else if ( rightHandSide.size ( ) == 2 ) {
		return addRule ( std::move ( leftHandSide ), std::make_pair ( std::move ( rightHandSide[0] ), std::move ( rightHandSide[1] ) ) );
	} else {
		throw GrammarException ( "Invalid right hand side" );
	}
}

template < class SymbolType >
std::map < SymbolType, std::set < std::vector < SymbolType > > > CNF < SymbolType >::getRawRules ( ) const {
	std::map < SymbolType, std::set < std::vector < SymbolType > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.template is < SymbolType > ( ) ) {
				std::vector < SymbolType > tmp { rhs.template get < SymbolType > ( ) };
				res[rule.first].insert ( std::move ( tmp ) );
			} else {
				const auto & realRHS = rhs.template get < std::pair < SymbolType, SymbolType > > ( );
				std::vector < SymbolType > tmp { realRHS.first, realRHS.second };
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	if ( getGeneratesEpsilon ( ) )
		res[getInitialSymbol ( )].insert ( std::vector < SymbolType > { } );

	return res;
}

template < class SymbolType >
bool CNF < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 ) {
		if ( leftHandSide != getInitialSymbol ( ) ) throw GrammarException ( "Illegal left hand side of epsilon rule" );

		bool res = getGeneratesEpsilon ( );
		setGeneratesEpsilon ( false );
		return res;
	} else if ( rightHandSide.size ( ) == 1 ) {
		return removeRule ( leftHandSide, rightHandSide[0] );
	} else if ( rightHandSide.size ( ) == 2 ) {
		return removeRule ( leftHandSide, std::make_pair ( rightHandSide[0], rightHandSide[1] ) );
	} else {
		throw GrammarException ( "Invalid right hand side" );
	}
}

template < class SymbolType >
void CNF < SymbolType >::setGeneratesEpsilon ( bool genEps ) {
	generatesEpsilon = genEps;
}

template < class SymbolType >
bool CNF < SymbolType >::getGeneratesEpsilon ( ) const {
	return generatesEpsilon;
}

template < class SymbolType >
int CNF < SymbolType >::compare ( const CNF & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void CNF < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(CNF "
	    << "nonterminalAlphabet = " << getNonterminalAlphabet ( )
	    << "terminalAlphabet = " << getTerminalAlphabet ( )
	    << "initialSymbol = " << getInitialSymbol ( )
	    << "rules = " << rules
	    << "generatesEpsilon = " << generatesEpsilon << ")";
}

template < class SymbolType >
CNF < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
CNF < SymbolType > CNF < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, CNF::getXmlTagName() );

	std::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	std::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	CNF < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	bool generatesEpsilon = GrammarFromXMLParser::parseGeneratesEpsilon ( input );
	grammar.setGeneratesEpsilon ( generatesEpsilon );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, CNF::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void CNF < SymbolType >::parseRule ( std::deque < sax::Token >::iterator & input, CNF & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > rhs = GrammarFromXMLParser::parseRuleOneOrTwoSymbolsRHS < SymbolType > ( input );

	grammar.addRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void CNF < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( CNF::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );
	GrammarToXMLComposer::composeGeneratesEpsilon ( out, this->getGeneratesEpsilon ( ) );

	out.emplace_back ( CNF::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void CNF < SymbolType >::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* CNF < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace alib {

template < class SymbolType >
class ComponentConstraint< grammar::CNF < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::CNF < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & rule : grammar.getRules ( ) )
			for ( const ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rhs : rule.second )
				if ( ( rhs.template is < SymbolType > ( ) && ( rhs.template get < SymbolType > ( ) == symbol ) ) || ( rhs.template get < std::pair < SymbolType, SymbolType > > ( ).first == symbol ) )
					return true;

		return false;
	}

	static bool available ( const grammar::CNF < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::CNF < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class ComponentConstraint< grammar::CNF < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::CNF < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const ext::variant < SymbolType, std::pair < SymbolType, SymbolType > > & rhs : rule.second )
				if ( rhs.template get < std::pair < SymbolType, SymbolType > > ( ).second == symbol )
					return true;

		}

		if ( grammar.template accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::CNF < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::CNF < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + ext::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::CNF < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::CNF < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::CNF < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* CNF_H_ */
