/*
 * GrammarFromXMLParser.h
 *
 *  Created on: Oct 12, 2013
 *      Author: Martin Zak
 */

#ifndef GRAMMAR_FROM_XML_PARSER_H_
#define GRAMMAR_FROM_XML_PARSER_H_

#include <deque>
#include <set>
#include <variant>
#include <vector>

#include <core/xmlApi.hpp>
#include <sax/ParserException.h>
#include <sax/FromXMLParserHelper.h>

namespace grammar {

/**
 * Parser used to get general FSM or EpsilonNFA, NFA, DFA from XML parsed into list of Tokens.
 */
class GrammarFromXMLParser {
public:
	template < class SymbolType >
	static std::set<SymbolType> parseNonterminalAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::set<SymbolType> parseTerminalAlphabet(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseInitialSymbol(std::deque<sax::Token>::iterator& input);

	static bool parseGeneratesEpsilon(std::deque<sax::Token>::iterator& input) {
		bool generatesEpsilon;
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "generatesEpsilon");
		if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "true")) {
			++input;
			sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "true");
			generatesEpsilon = true;
		} else {
			sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "false");
			sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "false");
			generatesEpsilon = false;
		}
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "generatesEpsilon");
		return generatesEpsilon;
	}

	template < class SymbolType >
	static std::vector<SymbolType> parseRuleLContext(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseRuleLHS(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseRuleRContext(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static SymbolType parseRuleSingleSymbolLHS(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::pair<SymbolType, std::vector<SymbolType>> parseRuleGNFRHS(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static std::vector<SymbolType> parseRuleRHS(std::deque<sax::Token>::iterator& input);
	template < class SymbolType >
	static ext::variant<SymbolType, std::pair<SymbolType, SymbolType>> parseRuleOneOrTwoSymbolsRHS(std::deque<sax::Token>::iterator& input);

	template<class T>
	static void parseRules(std::deque<sax::Token>::iterator& input, T& grammar);
};

template < class SymbolType >
std::set<SymbolType> GrammarFromXMLParser::parseNonterminalAlphabet(std::deque<sax::Token>::iterator& input) {
	std::set<SymbolType> inputSymbols;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "nonterminalAlphabet");
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		inputSymbols.insert(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "nonterminalAlphabet");
	return inputSymbols;
}

template < class SymbolType >
std::set<SymbolType> GrammarFromXMLParser::parseTerminalAlphabet(std::deque<sax::Token>::iterator& input) {
	std::set<SymbolType> inputSymbols;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "terminalAlphabet");
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		inputSymbols.insert(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "terminalAlphabet");
	return inputSymbols;
}

template < class SymbolType >
SymbolType GrammarFromXMLParser::parseInitialSymbol(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "initialSymbol");
	SymbolType blank(alib::xmlApi<SymbolType>::parse(input));
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "initialSymbol");
	return blank;
}

template < class SymbolType >
std::vector<SymbolType> GrammarFromXMLParser::parseRuleLContext(std::deque<sax::Token>::iterator& input) {
	std::vector<SymbolType> lContext;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "lContext");
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		++input;
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	} else while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		lContext.push_back(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "lContext");
	return lContext;
}

template < class SymbolType >
std::vector<SymbolType> GrammarFromXMLParser::parseRuleLHS(std::deque<sax::Token>::iterator& input) {
	std::vector<SymbolType> lhs;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "lhs");
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		++input;
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	} else while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		lhs.push_back(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "lhs");
	return lhs;
}

template < class SymbolType >
SymbolType GrammarFromXMLParser::parseRuleSingleSymbolLHS(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "lhs");
	SymbolType lhs = alib::xmlApi<SymbolType>::parse(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "lhs");
	return lhs;
}

template < class SymbolType >
std::vector<SymbolType> GrammarFromXMLParser::parseRuleRContext(std::deque<sax::Token>::iterator& input) {
	std::vector<SymbolType> rContext;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rContext");
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		++input;
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	} else while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		rContext.push_back(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rContext");
	return rContext;
}

template < class SymbolType >
std::pair<SymbolType, std::vector<SymbolType>> GrammarFromXMLParser::parseRuleGNFRHS(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rhs");
	SymbolType first = alib::xmlApi<SymbolType>::parse(input);
	std::vector<SymbolType> second;
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		second.push_back(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rhs");
	return std::make_pair(std::move(first), std::move(second));
}

template < class SymbolType >
std::vector<SymbolType> GrammarFromXMLParser::parseRuleRHS(std::deque<sax::Token>::iterator& input) {
	std::vector<SymbolType> rhs;
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rhs");
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "epsilon")) {
		++input;
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "epsilon");
	} else while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		rhs.push_back(alib::xmlApi<SymbolType>::parse(input));
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rhs");
	return rhs;
}

template < class SymbolType >
ext::variant<SymbolType, std::pair<SymbolType, SymbolType>> GrammarFromXMLParser::parseRuleOneOrTwoSymbolsRHS(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rhs");
	SymbolType first = alib::xmlApi<SymbolType>::parse(input);
	if(sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		SymbolType second = alib::xmlApi<SymbolType>::parse(input);
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rhs");
		return ext::variant<SymbolType, std::pair<SymbolType, SymbolType>>(std::make_pair(std::move(first), std::move(second)));
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rhs");
		return ext::variant<SymbolType, std::pair<SymbolType, SymbolType>>(std::move(first));
	}
}

template<class T>
void GrammarFromXMLParser::parseRules(std::deque<sax::Token>::iterator& input, T& grammar) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rules");
	while (sax::FromXMLParserHelper::isTokenType(input, sax::Token::TokenType::START_ELEMENT)) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "rule");
		T::parseRule(input, grammar);
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rule");
	}
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "rules");
}

} /* namespace grammar */

#endif /* GRAMMAR_FROM_XML_PARSER_H_ */

