/*
 * GrammarToXMLComposer.h
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#ifndef GRAMMAR_TO_XML_COMPOSER_H_
#define GRAMMAR_TO_XML_COMPOSER_H_

#include <deque>
#include <vector>
#include <tuple>
#include <set>
#include <variant>

#include <sax/Token.h>
#include <core/xmlApi.hpp>

namespace grammar {

/**
 * This class contains methods to print XML representation of automata to the output stream.
 */
class GrammarToXMLComposer {
public:
	template < class SymbolType >
	static void composeNonterminalAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols);
	template < class SymbolType >
	static void composeTerminalAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols);
	template < class SymbolType >
	static void composeInitialSymbol(std::deque<sax::Token>& out, const SymbolType& symbol);

	static void composeGeneratesEpsilon(std::deque<sax::Token>& out, bool generatesEpsilon) {
		out.emplace_back("generatesEpsilon", sax::Token::TokenType::START_ELEMENT);
		if(generatesEpsilon) {
			out.emplace_back("true", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("true", sax::Token::TokenType::END_ELEMENT);
		} else {
			out.emplace_back("false", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("false", sax::Token::TokenType::END_ELEMENT);
		}
		out.emplace_back("generatesEpsilon", sax::Token::TokenType::END_ELEMENT);
	}

	template < class SymbolType >
	static void composeRuleLContext(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleLHS(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleRContext(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleSingleSymbolLHS(std::deque<sax::Token>& out, const SymbolType& symbol);
	template < class SymbolType >
	static void composeRuleRHS(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols);
	template < class SymbolType >
	static void composeRuleOneOrTwoSymbolsRHS(std::deque<sax::Token>& out, const ext::variant<SymbolType, std::pair<SymbolType, SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleGNFRHS(std::deque<sax::Token>& out, const std::pair<SymbolType, std::vector<SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleLeftLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::pair<SymbolType, std::vector<SymbolType>>>& symbols);
	template < class SymbolType >
	static void composeRuleRightLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::pair<std::vector<SymbolType>, SymbolType>>& symbols);
	template < class SymbolType >
	static void composeRuleLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::tuple<std::vector<SymbolType>, SymbolType, std::vector<SymbolType>>>& symbols);
};

template < class SymbolType >
void GrammarToXMLComposer::composeNonterminalAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("nonterminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeTerminalAlphabet(std::deque<sax::Token>& out, const std::set<SymbolType>& symbols) {
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::START_ELEMENT);
	for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("terminalAlphabet", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeInitialSymbol(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("initialSymbol", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("initialSymbol", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLContext(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("lContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("lContext", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLHS(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRContext(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("rContext", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("rContext", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleSingleSymbolLHS(std::deque<sax::Token>& out, const SymbolType& symbol) {
	out.emplace_back("lhs", sax::Token::TokenType::START_ELEMENT);
	alib::xmlApi<SymbolType>::compose(out, symbol);
	out.emplace_back("lhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRHS(std::deque<sax::Token>& out, const std::vector<SymbolType>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.size() == 0) {
		out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
		out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
	} else for (const auto& symbol : symbols) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleOneOrTwoSymbolsRHS(std::deque<sax::Token>& out, const ext::variant<SymbolType, std::pair<SymbolType, SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<SymbolType>()) {
		alib::xmlApi<SymbolType>::compose(out, symbols.template get<SymbolType>());
	} else {
		const std::pair<SymbolType, SymbolType>& rhs = symbols.template get<std::pair<SymbolType, SymbolType>>();
		alib::xmlApi<SymbolType>::compose(out, rhs.first);
		alib::xmlApi<SymbolType>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleGNFRHS(std::deque<sax::Token>& out, const std::pair<SymbolType, std::vector<SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);

	alib::xmlApi<SymbolType>::compose(out, symbols.first);

	for (const auto& symbol : symbols.second) {
		alib::xmlApi<SymbolType>::compose(out, symbol);
	}

	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::tuple<std::vector<SymbolType>, SymbolType, std::vector<SymbolType>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<std::vector<SymbolType>>()) {
		const std::vector<SymbolType>& rhs = symbols.template get<std::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.template get<std::vector<SymbolType>>()) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
	} else {
		const std::tuple<std::vector<SymbolType>, SymbolType, std::vector<SymbolType>>& rhs = symbols.template get<std::tuple<std::vector<SymbolType>, SymbolType, std::vector<SymbolType>>>();
		for (const auto& symbol : std::get<0>(rhs)) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
		alib::xmlApi<SymbolType>::compose(out, std::get<1>(rhs));
		for (const auto& symbol : std::get<2>(rhs)) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleLeftLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::pair<SymbolType, std::vector<SymbolType>>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<std::vector<SymbolType>>()) {
		const std::vector<SymbolType>& rhs = symbols.template get<std::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else for (const auto& symbol : symbols.template get<std::vector<SymbolType>>()) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
	} else {
		const std::pair<SymbolType, std::vector<SymbolType>>& rhs = symbols.template get<std::pair<SymbolType, std::vector<SymbolType>>>();
		alib::xmlApi<SymbolType>::compose(out, rhs.first);
		for (const auto& symbol : rhs.second) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
void GrammarToXMLComposer::composeRuleRightLGRHS(std::deque<sax::Token>& out, const ext::variant<std::vector<SymbolType>, std::pair<std::vector<SymbolType>, SymbolType>>& symbols) {
	out.emplace_back("rhs", sax::Token::TokenType::START_ELEMENT);
	if(symbols.template is<std::vector<SymbolType>>()) {
		const std::vector<SymbolType>& rhs = symbols.template get<std::vector<SymbolType>>();
		if(rhs.size() == 0) {
			out.emplace_back("epsilon", sax::Token::TokenType::START_ELEMENT);
			out.emplace_back("epsilon", sax::Token::TokenType::END_ELEMENT);
		} else {
			for (const auto& symbol : symbols.template get<std::vector<SymbolType>>()) {
				alib::xmlApi<SymbolType>::compose(out, symbol);
			}
		}
	} else {
		const std::pair<std::vector<SymbolType>, SymbolType>& rhs = symbols.template get<std::pair<std::vector<SymbolType>, SymbolType>>();
		for (const auto& symbol : rhs.first) {
			alib::xmlApi<SymbolType>::compose(out, symbol);
		}
		alib::xmlApi<SymbolType>::compose(out, rhs.second);
	}
	out.emplace_back("rhs", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace grammar */

#endif /* GRAMMAR_TO_XML_COMPOSER_H_ */

