/*
 * FullAndLinearIndex.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_FULL_AND_LINEAR_INDEX_H_
#define ARBOLOGY_FULL_AND_LINEAR_INDEX_H_

#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsVector.h>

#include <indexes/stringology/PositionHeap.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class FullAndLinearIndex final : public alib::ObjectBase {
protected:
	indexes::stringology::PositionHeap < SymbolType > m_StringIndex;
	std::vector < int > m_JumpTable;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit FullAndLinearIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex, std::vector < int > jumpTable );

	/**
	 * @return Root node of the trie
	 */
	const indexes::stringology::PositionHeap < SymbolType > & getStringIndex ( ) const;

	/**
	 * @return subtree jump table
	 */
	const std::vector < int > & getJumps ( ) const;

	const std::vector < SymbolType > & getString ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return m_StringIndex.getAlphabet ( );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setStringIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return m_StringIndex.removeSymbolFromAlphabet ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FullAndLinearIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "FullAndLinearIndex";

		return xmlTagName;
	}

	static FullAndLinearIndex parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef FullAndLinearIndex < > normalized_type;

	virtual ObjectBase * normalize ( ) && {
		if ( typeid ( FullAndLinearIndex < > ) == typeid ( FullAndLinearIndex < SymbolType > ) )
			return this;

		indexes::stringology::PositionHeap < DefaultSymbolType > stringIndex = std::move ( m_StringIndex ).normalizeRaw ( );

		return new FullAndLinearIndex < > ( std::move ( stringIndex ), std::move ( m_JumpTable ) );
	}
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType >
FullAndLinearIndex < SymbolType >::FullAndLinearIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex, std::vector < int > jumpTable ) : m_StringIndex ( std::move ( stringIndex ) ), m_JumpTable ( jumpTable ) {
}

template < class SymbolType >
alib::ObjectBase * FullAndLinearIndex < SymbolType >::clone ( ) const {
	return new FullAndLinearIndex ( * this );
}

template < class SymbolType >
alib::ObjectBase * FullAndLinearIndex < SymbolType >::plunder ( ) && {
	return new FullAndLinearIndex ( std::move ( * this ) );
}

template < class SymbolType >
const indexes::stringology::PositionHeap < SymbolType > & FullAndLinearIndex < SymbolType >::getStringIndex ( ) const {
	return m_StringIndex;
}

template < class SymbolType >
const std::vector < int > & FullAndLinearIndex < SymbolType >::getJumps ( ) const {
	return m_JumpTable;
}

template < class SymbolType >
const std::vector < SymbolType > & FullAndLinearIndex < SymbolType >::getString ( ) const {
	return m_StringIndex.getString ( );
}

template < class SymbolType >
void FullAndLinearIndex < SymbolType >::setStringIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex ) {
	this->m_StringIndex = std::move ( stringIndex );
}

template < class SymbolType >
void FullAndLinearIndex < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FullAndLinearIndex " << this->m_StringIndex << ", " << m_JumpTable << ")";
}

template < class SymbolType >
int FullAndLinearIndex < SymbolType >::compare ( const FullAndLinearIndex & other ) const {
	auto first = std::tie ( getStringIndex ( ), getJumps ( ) );
	auto second = std::tie ( other.getStringIndex ( ), other.getJumps ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
FullAndLinearIndex < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
FullAndLinearIndex < SymbolType > FullAndLinearIndex < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, FullAndLinearIndex::getXmlTagName ( ) );
	indexes::stringology::PositionHeap < SymbolType > stringIndex = alib::xmlApi < indexes::stringology::PositionHeap < SymbolType > >::parse ( input );
	std::vector < int > jumps = alib::xmlApi < std::vector < int > >::parse ( input );

	FullAndLinearIndex < SymbolType > res ( std::move ( stringIndex ), std::move ( jumps ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, FullAndLinearIndex::getXmlTagName ( ) );
	return res;
}

template < class SymbolType >
void FullAndLinearIndex < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( FullAndLinearIndex::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < indexes::stringology::PositionHeap < SymbolType > >::compose ( out, getStringIndex ( ) );
	alib::xmlApi < std::vector < int > >::compose ( out, getJumps ( ) );
	out.emplace_back ( FullAndLinearIndex::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase * FullAndLinearIndex < SymbolType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

#endif /* ARBOLOGY_FULL_AND_LINEAR_INDEX_H_ */
