/*
 * SuffixArray.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_ARRAY_H_
#define SUFFIX_ARRAY_H_

#include <string>
#include <set>
#include <trie>
#include <variant>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>
#include <primitive/Unsigned.h>

#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class SuffixArray final : public alib::ObjectBase, public alib::Components < SuffixArray < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	std::vector < unsigned > m_data;
	std::vector < SymbolType > m_string;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixArray ( std::set < SymbolType > alphabet, std::vector < unsigned > data, std::vector < SymbolType > string );

	/**
	 * @return Root node of the trie
	 */
	const std::vector < unsigned > & getData ( ) const;

	const std::vector < SymbolType > & getString ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setData ( std::vector < unsigned > data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixArray & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "SuffixArray";

		return xmlTagName;
	}

	static SuffixArray parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef SuffixArray < > normalized_type;

	virtual ObjectBase * normalize ( ) && {
		if ( typeid ( SuffixArray < > ) == typeid ( SuffixArray < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( m_string ) );

		return new SuffixArray ( std::move ( alphabet ), std::move ( m_data ), std::move ( string ) );
	}
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
SuffixArray < SymbolType >::SuffixArray ( std::set < SymbolType > alphabet, std::vector < unsigned > data, std::vector < SymbolType > string ) : alib::Components < SuffixArray, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), m_data ( std::move ( data ) ), m_string ( std::move ( string ) ) {
	// TODO check validity of the string like in LinearString
}

template < class SymbolType >
alib::ObjectBase * SuffixArray < SymbolType >::clone ( ) const {
	return new SuffixArray ( * this );
}

template < class SymbolType >
alib::ObjectBase * SuffixArray < SymbolType >::plunder ( ) && {
	return new SuffixArray ( std::move ( * this ) );
}

template < class SymbolType >
const std::vector < unsigned > & SuffixArray < SymbolType >::getData ( ) const {
	return m_data;
}

template < class SymbolType >
const std::vector < SymbolType > & SuffixArray < SymbolType >::getString ( ) const {
	return m_string;
}

template < class SymbolType >
void SuffixArray < SymbolType >::setData ( std::vector < unsigned > data ) {
	this->m_data = std::move ( data );
}

template < class SymbolType >
void SuffixArray < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(SuffixArray " << this->m_data << ", " << this->m_string << ")";
}

template < class SymbolType >
int SuffixArray < SymbolType >::compare ( const SuffixArray & other ) const {
	auto first = std::tie ( getData ( ), getString ( ), getAlphabet ( ) );
	auto second = std::tie ( other.getData ( ), other.getString ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
SuffixArray < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
SuffixArray < SymbolType > SuffixArray < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixArray::getXmlTagName() );
	std::set < SymbolType > alphabet = alib::xmlApi < std::set < SymbolType > >::parse ( input );
	std::vector < unsigned > data = alib::xmlApi < std::vector < unsigned > >::parse ( input );
	std::vector < SymbolType > string = alib::xmlApi < std::vector < SymbolType > >::parse ( input );
	SuffixArray < SymbolType > res ( std::move ( alphabet ), std::move ( data ), std::move ( string ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixArray::getXmlTagName() );
	return res;
}

template < class SymbolType >
void SuffixArray < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixArray::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < std::set < SymbolType > >::compose ( out, getAlphabet ( ) );
	alib::xmlApi < std::vector < unsigned > >::compose ( out, getData ( ) );
	alib::xmlApi < std::vector < SymbolType > >::compose ( out, getString ( ) );
	out.emplace_back ( SuffixArray::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* SuffixArray < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType >
class ComponentConstraint < indexes::stringology::SuffixArray < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {
public:
	static bool used ( const indexes::stringology::SuffixArray < SymbolType > & index, const SymbolType & symbol ) {
		const std::vector < SymbolType > & content = index.getString ( );
		return std::find ( content.begin(), content.end(), symbol ) != content.end();
	}

	static bool available ( const indexes::stringology::SuffixArray < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::SuffixArray < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* SUFFIX_ARRAY_H_ */
