/*
 * SuffixTrie.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef SUFFIX_TRIE_H_
#define SUFFIX_TRIE_H_

#include <string>
#include <set>
#include <trie>
#include <variant>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/Object.h>
#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsTrie.h>
#include <container/ObjectsVariant.h>
#include <object/Void.h>

#include <primitive/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>
#include <indexes/common/IndexesNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class SuffixTrie final : public alib::ObjectBase, public alib::Components < SuffixTrie < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	ext::trie < SymbolType, ext::variant < void, unsigned > > m_trie;

public:
	/**
	 * @copydoc SuffixTrie::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrie::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit SuffixTrie ( std::set < SymbolType > edgeAlphabet );
	explicit SuffixTrie ( std::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, ext::variant < void, unsigned > > trie );
	explicit SuffixTrie ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie );

	void checkTrie ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie );

	/**
	 * @return Root node of the trie
	 */
	const ext::trie < SymbolType, ext::variant < void, unsigned > > & getRoot ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( ext::trie < SymbolType, ext::variant < void, unsigned > > tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromEdgeAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const SuffixTrie & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "SuffixTrie";

		return xmlTagName;
	}

	static SuffixTrie parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef SuffixTrie < > normalized_type;

	virtual alib::ObjectBase * normalize ( ) && {
		if ( typeid ( SuffixTrie < > ) == typeid ( SuffixTrie < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		ext::trie < DefaultSymbolType, ext::variant < void, unsigned > > trie = IndexesNormalize::normalizeTrie ( std::move ( m_trie ) );

		return new SuffixTrie < > ( std::move ( alphabet ), std::move ( trie ) );
	}
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( std::set < SymbolType > edgeAlphabet ) : SuffixTrie ( std::move ( edgeAlphabet ), ext::trie < SymbolType, ext::variant < void, unsigned > > ( ext::variant < void, unsigned >::template from < void > ( ) ) ) {
}

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( std::set < SymbolType > edgeAlphabet, ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) : alib::Components < SuffixTrie, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( edgeAlphabet ) ), std::tuple < > ( ) ), m_trie ( std::move ( trie ) ) {
	checkTrie ( this->m_trie );
}

template < class SymbolType >
SuffixTrie < SymbolType >::SuffixTrie ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) : SuffixTrie ( computeMinimalEdgeAlphabet ( trie ), trie ) {
}

template < class SymbolType >
alib::ObjectBase * SuffixTrie < SymbolType >::clone ( ) const {
	return new SuffixTrie ( * this );
}

template < class SymbolType >
alib::ObjectBase * SuffixTrie < SymbolType >::plunder ( ) && {
	return new SuffixTrie ( std::move ( * this ) );
}

template < class SymbolType >
void SuffixTrie < SymbolType >::checkTrie ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie ) {
	for ( const std::pair < const SymbolType, ext::trie < SymbolType, ext::variant < void, unsigned > > > & child : trie.getChildren ( ) ) {
		if ( ! getAlphabet ( ).count ( child.first ) )
			throw exception::CommonException ( "Symbol " + ext::to_string ( child.first ) + "not in the alphabet." );
		checkTrie ( child.second );
	}
}

template < class SymbolType >
const ext::trie < SymbolType, ext::variant < void, unsigned > > & SuffixTrie < SymbolType >::getRoot ( ) const {
	return m_trie;
}

template < class SymbolType >
void SuffixTrie < SymbolType >::setTree ( ext::trie < SymbolType, ext::variant < void, unsigned > > trie ) {
	checkTrie ( trie );
	this->m_trie = std::move ( trie ).plunder ( );
}

template < class SymbolType >
void SuffixTrie < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(SuffixTrie " << this->m_trie << ")";
}

template < class SymbolType >
int SuffixTrie < SymbolType >::compare ( const SuffixTrie & other ) const {
	auto first = std::tie ( getRoot ( ), getAlphabet ( ) );
	auto second = std::tie ( other.getRoot ( ), other.getAlphabet ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
SuffixTrie < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
SuffixTrie < SymbolType > SuffixTrie < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrie::getXmlTagName() );
	std::set < SymbolType > edgeAlphabet = alib::xmlApi < std::set < SymbolType > >::parse ( input );
	ext::trie < SymbolType, ext::variant < void, unsigned > > root = alib::xmlApi < ext::trie < SymbolType, ext::variant < void, unsigned > > >::parse ( input );
	SuffixTrie < SymbolType > trie ( std::move ( edgeAlphabet ), std::move ( root ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrie::getXmlTagName() );
	return trie;
}

template < class SymbolType >
void SuffixTrie < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrie::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < std::set < SymbolType > >::compose ( out, getAlphabet ( ) );
	alib::xmlApi < ext::trie < SymbolType, ext::variant < void, unsigned > > >::compose ( out, getRoot ( ) );
	out.emplace_back ( SuffixTrie::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* SuffixTrie < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace alib {

template < class SymbolType >
class ComponentConstraint < indexes::stringology::SuffixTrie < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {

	static bool used ( const ext::trie < SymbolType, ext::variant < void, unsigned > > & trie, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, ext::trie < SymbolType, ext::variant < void, unsigned > > > & child : trie.getChildren ( ) ) {
			if ( symbol == child.first || checkTrie ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::stringology::SuffixTrie < SymbolType > & index, const SymbolType & symbol ) {
		return used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::stringology::SuffixTrie < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::SuffixTrie < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* SUFFIX_TRIE_H_ */
