/*
 * FormalRegExpConcatenation.h
 *
 *  Created on: Nov 27, 2013
 *      Author: Martin Zak
 */

#ifndef FORMAL_REG_EXP_CONCATENATION_H_
#define FORMAL_REG_EXP_CONCATENATION_H_

#include <sstream>
#include <utility>

#include <exception/CommonException.h>

#include "FormalRegExpElement.h"

namespace regexp {

/**
 * Represents concatenation operator in the regular expression. Contains list of FormalRegExpElement
 * as operands of the operator.
 */
template < class SymbolType >
class FormalRegExpConcatenation : public FormalRegExpElement < SymbolType >, public ext::BinaryNode < std::smart_ptr < FormalRegExpElement < SymbolType > >, std::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpConcatenation < SymbolType > > {
public:
	void accept ( typename FormalRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit FormalRegExpConcatenation ( FormalRegExpElement < SymbolType > && left, FormalRegExpElement < SymbolType > && right );
	explicit FormalRegExpConcatenation ( const FormalRegExpElement < SymbolType > & left, const FormalRegExpElement < SymbolType > & right );

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	FormalRegExpConcatenation < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	FormalRegExpConcatenation < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	std::smart_ptr < UnboundedRegExpElement < SymbolType > > asUnbounded ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc FormalRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc FormalRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const override;

	/**
	 * @return elements
	 */
	const FormalRegExpElement < SymbolType > & getLeftElement ( ) const;
	const FormalRegExpElement < SymbolType > & getRightElement ( ) const;

	/**
	 * @return elements
	 */
	FormalRegExpElement < SymbolType > & getLeftElement ( );
	FormalRegExpElement < SymbolType > & getRightElement ( );

	/**
	 * @param element to append
	 */
	void setLeftElement ( FormalRegExpElement < SymbolType > && element );
	void setLeftElement ( const FormalRegExpElement < SymbolType > & element );

	void setRightElement ( FormalRegExpElement < SymbolType > && element );
	void setRightElement ( const FormalRegExpElement < SymbolType > & element );

	int compare ( const FormalRegExpElement < SymbolType > & other ) const override {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	int compare ( const FormalRegExpConcatenation & ) const;

	/**
	 * @copydoc FormalRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	explicit operator std::string ( ) const override;

	virtual std::smart_ptr < FormalRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		return std::smart_ptr < FormalRegExpElement < DefaultSymbolType > > ( new FormalRegExpConcatenation < DefaultSymbolType > ( std::move ( * std::move ( getLeftElement ( ) ).normalize ( ) ), std::move ( * std::move ( getRightElement ( ) ).normalize ( ) ) ) );
	}
};

} /* namespace regexp */

#include "../unbounded/UnboundedRegExpConcatenation.h"

namespace regexp {

template < class SymbolType >
FormalRegExpConcatenation < SymbolType >::FormalRegExpConcatenation ( FormalRegExpElement < SymbolType > && leftElement, FormalRegExpElement < SymbolType > && rightElement ) : ext::BinaryNode < std::smart_ptr < FormalRegExpElement < SymbolType > >, std::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpConcatenation < SymbolType > > ( std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( leftElement ).plunder ( ) ), std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( rightElement ).plunder ( ) ) ) {
}

template < class SymbolType >
FormalRegExpConcatenation < SymbolType >::FormalRegExpConcatenation ( const FormalRegExpElement < SymbolType > & left, const FormalRegExpElement < SymbolType > & right ) : FormalRegExpConcatenation ( std::move_copy ( left ), std::move_copy ( right ) ) {
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpConcatenation < SymbolType >::getLeftElement ( ) const {
	return * this->getLeft ( );
}

template < class SymbolType >
const FormalRegExpElement < SymbolType > & FormalRegExpConcatenation < SymbolType >::getRightElement ( ) const {
	return * this->getRight ( );
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpConcatenation < SymbolType >::getLeftElement ( ) {
	return * this->getLeft ( );
}

template < class SymbolType >
FormalRegExpElement < SymbolType > & FormalRegExpConcatenation < SymbolType >::getRightElement ( ) {
	return * this->getRight ( );
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::setLeftElement ( FormalRegExpElement < SymbolType > && element ) {
	this->setLeft( std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::setLeftElement ( const FormalRegExpElement < SymbolType > & element ) {
	setLeftElement ( std::move_copy ( element ) );
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::setRightElement ( FormalRegExpElement < SymbolType > && element ) {
	this->setRight ( std::smart_ptr < FormalRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::setRightElement ( const FormalRegExpElement < SymbolType > & element ) {
	setRightElement ( std::move_copy ( element ) );
}

template < class SymbolType >
FormalRegExpConcatenation < SymbolType > * FormalRegExpConcatenation < SymbolType >::clone ( ) const {
	return new FormalRegExpConcatenation ( * this );
}

template < class SymbolType >
FormalRegExpConcatenation < SymbolType > * FormalRegExpConcatenation < SymbolType >::plunder ( ) && {
	return new FormalRegExpConcatenation ( std::move ( * this ) );
}

template < class SymbolType >
std::smart_ptr < UnboundedRegExpElement < SymbolType > > FormalRegExpConcatenation < SymbolType >::asUnbounded ( ) const {
	UnboundedRegExpConcatenation < SymbolType > * res = new UnboundedRegExpConcatenation < SymbolType > ( );

	res->appendElement ( std::move ( * getLeftElement ( ).asUnbounded ( ) ) );
	res->appendElement ( std::move ( * getRightElement ( ).asUnbounded ( ) ) );

	return std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( res );
}

template < class SymbolType >
int FormalRegExpConcatenation < SymbolType >::compare ( const FormalRegExpConcatenation < SymbolType > & other ) const {
	int res = getLeftElement ( ).compare ( other.getLeftElement ( ) );

	if ( res == 0 ) res = getRightElement ( ).compare ( other.getRightElement ( ) );

	return res;
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FormalRegExpConcatenation";
	out << " " << getLeftElement ( );
	out << " " << getRightElement ( );
	out << ")";
}

template < class SymbolType >
bool FormalRegExpConcatenation < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	return getLeftElement ( ).testSymbol ( symbol ) && this->getRight ( )->testSymbol ( symbol );
}

template < class SymbolType >
void FormalRegExpConcatenation < SymbolType >::computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const {
	getLeftElement ( ).computeMinimalAlphabet ( alphabet );
	getRightElement ( ).computeMinimalAlphabet ( alphabet );
}

template < class SymbolType >
bool FormalRegExpConcatenation < SymbolType >::checkAlphabet ( const std::set < SymbolType > & alphabet ) const {
	return getLeftElement ( ).checkAlphabet ( alphabet ) && getRightElement ( ).checkAlphabet ( alphabet );
}

template < class SymbolType >
FormalRegExpConcatenation < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_CONCATENATION_H_ */
