/*
 * UnboundedRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_ELEMENT_H_
#define UNBOUNDED_REG_EXP_ELEMENT_H_

#include <set>
#include <tree>
#include <core/visitor.hpp>
#include "../RegExpFeatures.h"

namespace regexp {

/**
 * Abstract class representing element in the regular expression. Can be operator or symbol.
 */
template < class SymbolType >
class UnboundedRegExpElement : public alib::CommonBaseMiddle < UnboundedRegExpElement < SymbolType > >, public ext::BaseNode < UnboundedRegExpElement < SymbolType > > {
public:
	class Visitor {
	public:
		virtual void visit ( const UnboundedRegExpAlternation < SymbolType > & ) = 0;
		virtual void visit ( const UnboundedRegExpConcatenation < SymbolType > & ) = 0;
		virtual void visit ( const UnboundedRegExpIteration < SymbolType > & ) = 0;
		virtual void visit ( const UnboundedRegExpSymbol < SymbolType > & ) = 0;
		virtual void visit ( const UnboundedRegExpEmpty < SymbolType > & ) = 0;
		virtual void visit ( const UnboundedRegExpEpsilon < SymbolType > & ) = 0;
	};

	template < class ReturnType, class Visitorr, class ... Params >
	class VisitorContext : public alib::VisitorContextAux < ReturnType, Visitorr, Params ... >, public UnboundedRegExpElement::Visitor {
	public:
		using alib::VisitorContextAux < ReturnType, Visitorr, Params ... >::VisitorContextAux;

		void visit ( const UnboundedRegExpAlternation < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const UnboundedRegExpConcatenation < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const UnboundedRegExpIteration < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const UnboundedRegExpSymbol < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const UnboundedRegExpEmpty < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}

		void visit ( const UnboundedRegExpEpsilon < SymbolType > & inherit ) override {
			this->call ( inherit, std::make_index_sequence < sizeof ... ( Params ) > { } );
		}
	};

	template < class ReturnType, class Visitorr, class ... Params >
	ReturnType accept ( Params && ... params ) const {
		VisitorContext < ReturnType, Visitorr, Params ... > context ( std::forward < Params > ( params ) ... );
		accept ( context );
		return context.getResult ( );
	}

	virtual void accept ( UnboundedRegExpElement::Visitor & visitor ) const = 0;

	/**
	 * Creates copy of the element.
	 * @return copy of the element
	 */
	virtual std::smart_ptr < FormalRegExpElement < SymbolType > > asFormal ( ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const SymbolType & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree and checks whether all symbols in the regexp tree are in the alphabet
	 *
	 * @param alphabet
	 * @return true if symbols in the regexp are in the alphabet, false otherwise
	 */
	virtual bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const = 0;

	/**
	 * @copydoc RankedNode::computeMinimalAlphabet()
	 */
	std::set < SymbolType > computeMinimalAlphabet ( ) const;

	/** Traverses the regexp tree and normalizes the symbols to DefaultSymbolType
	 */
	virtual std::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > normalize ( ) && = 0;
};

template < class SymbolType >
std::set < SymbolType > UnboundedRegExpElement < SymbolType >::computeMinimalAlphabet ( ) const {
	std::set < SymbolType > res;

	computeMinimalAlphabet ( res );
	return res;
}

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_ELEMENT_H_ */
