/*
 * UnboundedRegExpIteration.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef UNBOUNDED_REG_EXP_ITERATION_H_
#define UNBOUNDED_REG_EXP_ITERATION_H_

#include <sstream>

#include <exception/CommonException.h>

#include "UnboundedRegExpElement.h"

namespace regexp {

/**
 * Represents iteration operator in the regular expression. Contains one UnboundedRegExpElement
 * as operand.
 */
template < class SymbolType >
class UnboundedRegExpIteration : public UnboundedRegExpElement < SymbolType >, public ext::UnaryNode < std::smart_ptr < UnboundedRegExpElement < SymbolType > >, std::smart_ptr < const UnboundedRegExpElement < SymbolType > >, UnboundedRegExpIteration < SymbolType > > {
public:
	void accept ( typename UnboundedRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit UnboundedRegExpIteration ( UnboundedRegExpElement < SymbolType > && );
	explicit UnboundedRegExpIteration ( const UnboundedRegExpElement < SymbolType > & );

	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	UnboundedRegExpIteration < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	UnboundedRegExpIteration < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc UnboundedRegExpElement::cloneAsFormal() const
	 */
	std::smart_ptr < FormalRegExpElement < SymbolType > > asFormal ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const override;

	/**
	 * @return element
	 */
	const UnboundedRegExpElement < SymbolType > & getElement ( ) const;

	/**
	 * @return element
	 */
	UnboundedRegExpElement < SymbolType > & getElement ( );

	/**
	 * @param element to iterate
	 */
	void setElement ( UnboundedRegExpElement < SymbolType > && element );
	void setElement ( const UnboundedRegExpElement < SymbolType > & element );

	int compare ( const UnboundedRegExpElement < SymbolType > & other ) const override {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	int compare ( const UnboundedRegExpIteration & ) const;

	/**
	 * @copydoc UnboundedRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	explicit operator std::string ( ) const override;

	virtual std::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		return std::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > ( new UnboundedRegExpIteration < DefaultSymbolType > ( std::move ( * std::move ( getElement ( ) ).normalize ( ) ) ) );
	}
};

} /* namespace regexp */

#include "UnboundedRegExpEmpty.h"
#include "../formal/FormalRegExpIteration.h"

namespace regexp {

template < class SymbolType >
UnboundedRegExpIteration < SymbolType >::UnboundedRegExpIteration ( UnboundedRegExpElement < SymbolType > && element ) : ext::UnaryNode < std::smart_ptr < UnboundedRegExpElement < SymbolType > >, std::smart_ptr < const UnboundedRegExpElement < SymbolType > >, UnboundedRegExpIteration < SymbolType > > ( std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) ) {
}

template < class SymbolType >
UnboundedRegExpIteration < SymbolType >::UnboundedRegExpIteration ( const UnboundedRegExpElement < SymbolType > & element ) : UnboundedRegExpIteration ( std::move_copy ( element ) ) {
}

template < class SymbolType >
const UnboundedRegExpElement < SymbolType > & UnboundedRegExpIteration < SymbolType >::getElement ( ) const {
	return * this->getChild ( );
}

template < class SymbolType >
UnboundedRegExpElement < SymbolType > & UnboundedRegExpIteration < SymbolType >::getElement ( ) {
	return * this->getChild ( );
}

template < class SymbolType >
void UnboundedRegExpIteration < SymbolType >::setElement ( UnboundedRegExpElement < SymbolType > && elementParam ) {
	this->setChild ( std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( std::move ( elementParam ).plunder ( ) ) );
}

template < class SymbolType >
void UnboundedRegExpIteration < SymbolType >::setElement ( const UnboundedRegExpElement < SymbolType > & elementParam ) {
	setElement ( std::move_copy ( elementParam ) );
}

template < class SymbolType >
UnboundedRegExpIteration < SymbolType > * UnboundedRegExpIteration < SymbolType >::clone ( ) const {
	return new UnboundedRegExpIteration ( * this );
}

template < class SymbolType >
UnboundedRegExpIteration < SymbolType > * UnboundedRegExpIteration < SymbolType >::plunder ( ) && {
	return new UnboundedRegExpIteration ( std::move ( * this ) );
}

template < class SymbolType >
std::smart_ptr < FormalRegExpElement < SymbolType > > UnboundedRegExpIteration < SymbolType >::asFormal ( ) const {
	return std::smart_ptr < FormalRegExpElement < SymbolType > > ( new FormalRegExpIteration < SymbolType > ( std::move ( * getElement ( ).asFormal ( ) ) ) );
}

template < class SymbolType >
int UnboundedRegExpIteration < SymbolType >::compare ( const UnboundedRegExpIteration < SymbolType > & other ) const {
	return getElement ( ).compare ( other.getElement ( ) );
}

template < class SymbolType >
void UnboundedRegExpIteration < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(RegExpUnboundedRegExpIteration " << getElement ( ) << ")";
}

template < class SymbolType >
bool UnboundedRegExpIteration < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	return getElement ( ).testSymbol ( symbol );
}

template < class SymbolType >
bool UnboundedRegExpIteration < SymbolType >::checkAlphabet ( const std::set < SymbolType > & alphabet ) const {
	return getElement ( ).checkAlphabet ( alphabet );
}

template < class SymbolType >
void UnboundedRegExpIteration < SymbolType >::computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const {
	getElement ( ).computeMinimalAlphabet ( alphabet );
}

template < class SymbolType >
UnboundedRegExpIteration < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_ITERATION_H_ */
