/*
 * FormalRTEStructure.h
 *
 *  Created on: Oct 20, 2016
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_RTE_STRUCTURE_H_
#define FORMAL_RTE_STRUCTURE_H_

#include <memory>
#include "FormalRTEElement.h"

#include "FormalRTEEmpty.h"

#include "../common/RTEFromXMLParser.h"
#include "../common/RTEToXMLComposer.h"
#include <exception/CommonException.h>

namespace rte {

/**
 * Represents formal regular expression structure as parsed from the XML. Regular expression is stored
 * as a tree of FormalRTEElement.
 */
template < class SymbolType, class RankType >
class FormalRTEStructure {
protected:
	std::smart_ptr < FormalRTEElement < SymbolType, RankType > > m_structure;

public:
	explicit FormalRTEStructure ( );

	explicit FormalRTEStructure ( FormalRTEElement < SymbolType, RankType > && structure );
	explicit FormalRTEStructure ( const FormalRTEElement < SymbolType, RankType > & structure );

	/**
	 * @return Root node of the formal regular expression tree
	 */
	const FormalRTEElement < SymbolType, RankType > & getStructure ( ) const;

	/**
	 * @return Root node of the formal regular expression tree
	 */
	FormalRTEElement < SymbolType, RankType > & getStructure ( );

	/**
	 * Sets the root node of the formal regular expression tree
	 * @param structure root node to set
	 */
	void setStructure ( FormalRTEElement < SymbolType, RankType > && structure );
	void setStructure ( const FormalRTEElement < SymbolType, RankType > & structure );

	friend std::ostream & operator << ( std::ostream & out, const FormalRTEStructure < SymbolType, RankType > & structure ) {
		out << structure.getStructure ( );
		return out;
	}

	friend bool operator < ( const FormalRTEStructure < SymbolType, RankType > & first, const FormalRTEStructure < SymbolType, RankType > & second ) {
		return first.getStructure().compare(second.getStructure()) < 0;
	}

	friend bool operator == ( const FormalRTEStructure < SymbolType, RankType > & first, const FormalRTEStructure < SymbolType, RankType > & second ) {
		return first.getStructure().compare(second.getStructure()) == 0;
	}

	FormalRTEStructure < DefaultSymbolType, DefaultRankType > normalize ( ) && {
		return FormalRTEStructure < DefaultSymbolType, DefaultRankType > ( std::move ( * std::move ( getStructure ( ) ).normalize ( ) ) );
	}

};

} /* namespace rte */

namespace alib {

template < class SymbolType, class RankType >
struct xmlApi < rte::FormalRTEStructure < SymbolType, RankType > > {
	static rte::FormalRTEStructure < SymbolType, RankType > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const rte::FormalRTEStructure < SymbolType, RankType > & data );
};

} /* namespace alib */

namespace ext {

template < class SymbolType, class RankType >
struct compare < rte::FormalRTEStructure < SymbolType, RankType > > {
	int operator()(const rte::FormalRTEStructure < SymbolType, RankType > & first, const rte::FormalRTEStructure < SymbolType, RankType > & second) const {
		return first.getStructure().compare(second.getStructure());
	}
};

} /* namespace ext */

namespace rte {

template < class SymbolType, class RankType >
FormalRTEStructure < SymbolType, RankType >::FormalRTEStructure ( FormalRTEElement < SymbolType, RankType > && structure ) : m_structure ( NULL ) {
	setStructure ( std::move ( structure ) );
}

template < class SymbolType, class RankType >
FormalRTEStructure < SymbolType, RankType >::FormalRTEStructure ( const FormalRTEElement < SymbolType, RankType > & structure ) : FormalRTEStructure ( std::move_copy ( structure ) ) {
}

template < class SymbolType, class RankType >
FormalRTEStructure < SymbolType, RankType >::FormalRTEStructure ( ) : FormalRTEStructure ( FormalRTEEmpty < SymbolType, RankType > ( ) ) {
}

template < class SymbolType, class RankType >
const FormalRTEElement < SymbolType, RankType > & FormalRTEStructure < SymbolType, RankType >::getStructure ( ) const {
	return * m_structure;
}

template < class SymbolType, class RankType >
FormalRTEElement < SymbolType, RankType > & FormalRTEStructure < SymbolType, RankType >::getStructure ( ) {
	return * m_structure;
}

template < class SymbolType, class RankType >
void FormalRTEStructure < SymbolType, RankType >::setStructure ( const FormalRTEElement < SymbolType, RankType > & structure ) {
	setStructure ( std::move_copy ( structure ) );
}

template < class SymbolType, class RankType >
void FormalRTEStructure < SymbolType, RankType >::setStructure ( FormalRTEElement < SymbolType, RankType > && param ) {
	this->m_structure = std::smart_ptr < FormalRTEElement < SymbolType, RankType > > ( std::move ( param ).plunder ( ) );
}

} /* namespace rte */

namespace alib {

template < class SymbolType, class RankType >
rte::FormalRTEStructure < SymbolType, RankType > xmlApi < rte::FormalRTEStructure < SymbolType, RankType > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return rte::FormalRTEStructure < SymbolType, RankType > ( rte::RTEFromXMLParser::parseFormalRTEElement < SymbolType, RankType > ( input ) );
}

template < class SymbolType, class RankType >
bool xmlApi < rte::FormalRTEStructure < SymbolType, RankType > >::first ( const std::deque < sax::Token >::const_iterator & ) {
	throw exception::CommonException ( "Unimplemented" );
}

template < class SymbolType, class RankType >
void xmlApi < rte::FormalRTEStructure < SymbolType, RankType > >::compose ( std::deque < sax::Token > & output, const rte::FormalRTEStructure < SymbolType, RankType > & data ) {
	data.getStructure ( ).template accept < void, rte::RTEToXMLComposer::Formal > ( output );
}

} /* namespace alib */

#endif /* FORMAL_RTE_STRUCTURE_H_ */
