/*
 * LinearString.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef LINEAR_STRING_H_
#define LINEAR_STRING_H_

#include <set>
#include <vector>
#include <sstream>
#include <algorithm>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "StringBase.h"
#include "StringFeatures.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include <exception/CommonException.h>
#include <alphabet/common/SymbolNormalize.h>

#include "../tree/TreeFeatures.h"

namespace string {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType >
class LinearString final : public StringBase, public alib::Components < LinearString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::vector < SymbolType > m_Data;

public:
	explicit LinearString ( );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedTree < TreeSymbolType, TreeRankType > & tree );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedPattern < TreeSymbolType, TreeRankType > & tree );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedNonlinearPattern < TreeSymbolType, TreeRankType > & tree );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedBarTree < TreeSymbolType, TreeRankType > & tree );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedBarPattern < TreeSymbolType, TreeRankType > & tree );
	template < class TreeSymbolType, class TreeRankType >
	explicit LinearString ( const tree::PrefixRankedBarNonlinearPattern < TreeSymbolType, TreeRankType > & tree );
	explicit LinearString ( const tree::PrefixBarTree < SymbolType > & tree );
	explicit LinearString ( std::set < SymbolType > alphabet, std::vector < SymbolType > str );
	explicit LinearString ( std::vector < SymbolType > str );
	explicit LinearString ( const std::string & str );
	explicit LinearString ( const char * str );
	explicit LinearString ( const Epsilon < SymbolType > & epsilon );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	/**
	 * @param element to append
	 */
	void appendSymbol ( SymbolType symbol );

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < SymbolType > & getContent ( ) const;

	void setContent ( std::vector < SymbolType > str );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LinearString & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LinearString";

		return xmlTagName;
	}

	static LinearString parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef LinearString < > normalized_type;

	virtual StringBase * normalize ( ) && {
		if ( typeid ( LinearString < > ) == typeid ( LinearString < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < DefaultSymbolType > content = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( m_Data ) );
		return new LinearString < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace string */

#include "Epsilon.h"
#include "../tree/ranked/PrefixRankedTree.h"
#include "../tree/ranked/PrefixRankedPattern.h"
#include "../tree/ranked/PrefixRankedNonlinearPattern.h"
#include "../tree/ranked/PrefixRankedBarTree.h"
#include "../tree/ranked/PrefixRankedBarPattern.h"
#include "../tree/ranked/PrefixRankedBarNonlinearPattern.h"
#include "../tree/unranked/PrefixBarTree.h"
#include "../tree/common/TreeAuxiliary.h"

namespace string {

template < class SymbolType >
LinearString < SymbolType >::LinearString(std::set<SymbolType> alphabet, std::vector<SymbolType> str) : alib::Components < LinearString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent(std::move(str));
}

template < class SymbolType >
LinearString < SymbolType >::LinearString() : LinearString ( std::set < SymbolType > ( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(std::vector<SymbolType> str) : LinearString ( std::set < SymbolType> ( str.begin(), str.end() ), str ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const std::string & str) : LinearString ( std::vector < SymbolType > ( str.begin ( ), str.end ( ) ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const char* str) : LinearString ( ( std::string ) str ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString(const Epsilon < SymbolType > & epsilon) : LinearString ( epsilon.getAlphabet( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedTree < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedPattern < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedNonlinearPattern < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarTree < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarPattern < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
template < class TreeSymbolType, class TreeRankType >
LinearString < SymbolType >::LinearString ( const tree::PrefixRankedBarNonlinearPattern < TreeSymbolType, TreeRankType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
LinearString < SymbolType >::LinearString ( const tree::PrefixBarTree < SymbolType > & tree ) : LinearString ( tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType >
StringBase * LinearString < SymbolType >::clone ( ) const {
	return new LinearString ( * this );
}

template < class SymbolType >
StringBase * LinearString < SymbolType >::plunder ( ) && {
	return new LinearString ( std::move ( * this ) );
}

template < class SymbolType >
void LinearString < SymbolType >::appendSymbol ( SymbolType symbol ) {
	if ( getAlphabet().count ( symbol ) == 0 )
		throw exception::CommonException ( "Input symbol \"" + ext::to_string ( symbol ) + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

template < class SymbolType >
const std::vector < SymbolType > & LinearString < SymbolType >::getContent ( ) const {
	return this->m_Data;
}

template < class SymbolType >
void LinearString < SymbolType >::setContent ( std::vector < SymbolType > str ) {
	std::set < SymbolType > minimalAlphabet ( str.begin ( ), str.end ( ) );
	std::set < SymbolType > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( str );
}

template < class SymbolType >
bool LinearString < SymbolType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType >
int LinearString < SymbolType >::compare ( const LinearString & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void LinearString < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(LinearString ";

	for ( const SymbolType & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType >
LinearString < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const SymbolType & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType >
LinearString < SymbolType > LinearString < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearString::getXmlTagName() );
	std::set < SymbolType > alphabet = StringFromXMLParser::parseAlphabet < SymbolType > ( input );
	std::vector < SymbolType > content = StringFromXMLParser::parseContent < SymbolType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearString::getXmlTagName() );

	return LinearString < SymbolType > ( alphabet, content );
}

template < class SymbolType >
void LinearString < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearString::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, getAlphabet() );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearString::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* LinearString < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace string */

namespace alib {

template < class SymbolType >
class ComponentConstraint< ::string::LinearString < SymbolType >, SymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::LinearString < SymbolType > & str, const SymbolType & symbol ) {
		const std::vector<SymbolType>& content = str.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::LinearString < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::string::LinearString < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace alib */

#endif /* LINEAR_STRING_H_ */
