/*
 * TreeFromXMLParser.h
 *
 *  Created on: Nov 16, 2014
 *      Author: Stepan Plachy
 */

#ifndef TREE_FROM_XML_PARSER_H_
#define TREE_FROM_XML_PARSER_H_

#include <set>
#include <vector>
#include <deque>
#include <tree>

#include <sax/Token.h>
#include <sax/ParserException.h>
#include <core/xmlApi.hpp>

namespace tree {

/**
 * Parser used to get tree from XML parsed into list of Tokens.
 */
class TreeFromXMLParser {
	template < class SymbolType >
	static ext::tree < SymbolType > parseTreeChild ( std::deque < sax::Token >::iterator & input );

public:
	template < class SymbolType >
	static std::vector < SymbolType > parseLinearContent ( std::deque < sax::Token >::iterator & input );
	template < class SymbolType >
	static ext::tree < SymbolType > parseTreeContent ( std::deque < sax::Token >::iterator & input );

	template < class SymbolType >
	static std::set < SymbolType > parseAlphabet ( std::deque < sax::Token >::iterator & input );

	template < class SymbolType >
	static SymbolType parseBar ( std::deque < sax::Token >::iterator & input );
	template < class SymbolType >
	static std::set < SymbolType > parseBars ( std::deque < sax::Token >::iterator & input );
	template < class SymbolType >
	static SymbolType parseVariablesBar ( std::deque < sax::Token >::iterator & input );

	template < class SymbolType >
	static SymbolType parseSubtreeWildcardSymbol ( std::deque < sax::Token >::iterator & input );

	template < class SymbolType >
	static std::set < SymbolType > parseNonlinearVariables ( std::deque < sax::Token >::iterator & input );
};

template < class SymbolType >
SymbolType TreeFromXMLParser::parseBar ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "bar" );
	SymbolType bar ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "bar" );

	return bar;
}

template < class SymbolType >
std::set < SymbolType > TreeFromXMLParser::parseBars ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > bars;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "bars" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		bars.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "bars" );

	return bars;
}

template < class SymbolType >
SymbolType TreeFromXMLParser::parseVariablesBar ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "variablesBar" );
	SymbolType variablesBar ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "variablesBar" );

	return variablesBar;
}

template < class SymbolType >
SymbolType TreeFromXMLParser::parseSubtreeWildcardSymbol ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "subtreeWildcard" );
	SymbolType subtreeWildcard ( alib::xmlApi < SymbolType >::parse ( input ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "subtreeWildcard" );

	return subtreeWildcard;
}

template < class SymbolType >
std::set < SymbolType > TreeFromXMLParser::parseAlphabet ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > symbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "alphabet" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		symbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "alphabet" );
	return symbols;
}

template < class SymbolType >
std::vector < SymbolType > TreeFromXMLParser::parseLinearContent ( std::deque < sax::Token >::iterator & input ) {
	std::vector < SymbolType > data;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "content" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		data.push_back ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "content" );
	return data;
}

template < class SymbolType >
ext::tree < SymbolType > TreeFromXMLParser::parseTreeChild ( std::deque < sax::Token >::iterator & input ) {
	SymbolType symbol = alib::xmlApi < SymbolType >::parse ( input );

	std::vector < ext::tree < SymbolType > > children;

	if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" ) ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" );

		while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) ) {
			children.push_back ( TreeFromXMLParser::parseTreeChild < SymbolType > ( input ) );
		}

		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" );
	}

	return ext::tree < SymbolType > ( symbol, std::move( children ) );
}

template < class SymbolType >
ext::tree < SymbolType > TreeFromXMLParser::parseTreeContent ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "content" );

	ext::tree < SymbolType > data = TreeFromXMLParser::parseTreeChild < SymbolType > ( input );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "content" );
	return data;
}

template < class SymbolType >
std::set < SymbolType > TreeFromXMLParser::parseNonlinearVariables ( std::deque < sax::Token >::iterator & input ) {
	std::set < SymbolType > symbols;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "nonlinearVariables" );

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		symbols.insert ( alib::xmlApi < SymbolType >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "nonlinearVariables" );
	return symbols;
}

} /* namespace tree */

#endif /* TREE_FROM_XML_PARSER_H_ */
