/*
 * PrefixRankedNonlinearPattern.h
 *
 *  Created on: Jul 20, 2016
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_NONLINEAR_PATTERN_H_
#define PREFIX_RANKED_NONLINEAR_PATTERN_H_

#include <set>
#include <vector>
#include <tree>
#include <sstream>
#include <algorithm>
#include <deque>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <common/ranked_symbol.hpp>

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../RankedTreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include <tree/common/TreeNormalize.h>

#include <alphabet/SubtreeWildcardSymbol.h>

namespace tree {

class GeneralAlphabet;
class SubtreeWildcard;
class NonlinearAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType, class RankType >
class PrefixRankedNonlinearPattern final : public RankedTreeBase, public alib::Components < PrefixRankedNonlinearPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > {
	std::vector < common::ranked_symbol < SymbolType, RankType > > m_Data;

	static std::vector < common::ranked_symbol < SymbolType, RankType > > toPrefixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree );

public:
	explicit PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::set < common::ranked_symbol < SymbolType, RankType > > nonlinearVariables, std::set < common::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::set < common::ranked_symbol < SymbolType, RankType > > nonlinearVariables, std::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedNonlinearPattern ( const PrefixRankedTree < SymbolType, RankType > & tree );
	explicit PrefixRankedNonlinearPattern ( const PrefixRankedPattern < SymbolType, RankType > & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedTree < SymbolType, RankType > & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedPattern < SymbolType, RankType > & tree );
	explicit PrefixRankedNonlinearPattern ( const RankedNonlinearPattern < SymbolType, RankType > & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < common::ranked_symbol < SymbolType, RankType > > & getContent ( ) const;

	const std::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < common::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const common::ranked_symbol < SymbolType, RankType > & getSubtreeWildcard ( ) const {
		return this->template accessElement < SubtreeWildcard > ( ).get ( );
	}

	const std::set < common::ranked_symbol < SymbolType, RankType > > & getNonlinearVariables ( ) const {
		return this->template accessComponent < NonlinearAlphabet > ( ).get ( );
	}

	void setContent ( std::vector < common::ranked_symbol < SymbolType, RankType > > data );

	void arityChecksum ( const std::vector < common::ranked_symbol < SymbolType, RankType > > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedNonlinearPattern & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "PrefixRankedNonlinearPattern";

		return xmlTagName;
	}

	static PrefixRankedNonlinearPattern parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef PrefixRankedNonlinearPattern < > normalized_type;

	virtual RankedTreeBase * normalize ( ) && {
		if ( typeid ( PrefixRankedNonlinearPattern < > ) == typeid ( PrefixRankedNonlinearPattern < SymbolType, RankType > ) )
			return this;

		common::ranked_symbol < DefaultSymbolType, DefaultRankType > wildcard = alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( this->template accessElement < SubtreeWildcard > ( ).get ( ) ) );
		std::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > nonlinearAlphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < NonlinearAlphabet > ( ).get ( ) ) );
		std::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = alphabet::SymbolNormalize::normalizeRankedSymbols ( std::move ( m_Data ) );

		return new PrefixRankedNonlinearPattern < > ( std::move ( wildcard ), std::move ( nonlinearAlphabet ), std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace tree */

#include "PrefixRankedTree.h"
#include "PrefixRankedPattern.h"
#include "RankedTree.h"
#include "RankedPattern.h"
#include "RankedNonlinearPattern.h"

namespace tree {

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::set < common::ranked_symbol < SymbolType, RankType > > nonlinearVariables, std::set < common::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < common::ranked_symbol < SymbolType, RankType > > data ) : alib::Components < PrefixRankedNonlinearPattern, common::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet, NonlinearAlphabet >, std::tuple < SubtreeWildcard > > ( std::make_tuple ( std::move ( alphabet ), std::move ( nonlinearVariables ) ), std::make_tuple ( subtreeWildcard ) ) {
	setContent ( std::move ( data ) );
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::set < common::ranked_symbol < SymbolType, RankType > > nonlinearVariables, std::vector < common::ranked_symbol < SymbolType, RankType > > data ) : PrefixRankedNonlinearPattern ( subtreeWildcard, nonlinearVariables, std::set < common::ranked_symbol < SymbolType, RankType > > ( data.begin ( ), data.end ( ) ) + std::set < common::ranked_symbol < SymbolType, RankType > > { subtreeWildcard } + nonlinearVariables, data ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( common::ranked_symbol < SymbolType, RankType > subtreeWildcard, std::vector < common::ranked_symbol < SymbolType, RankType > > data ) : PrefixRankedNonlinearPattern ( subtreeWildcard, {}, std::set < common::ranked_symbol < SymbolType, RankType > > ( data.begin ( ), data.end ( ) ) + std::set < common::ranked_symbol < SymbolType, RankType > > { subtreeWildcard }, data ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( const PrefixRankedTree < SymbolType, RankType > & tree ) : PrefixRankedNonlinearPattern ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < SymbolType, RankType > > ( ), { }, tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( const PrefixRankedPattern < SymbolType, RankType > & tree ) : PrefixRankedNonlinearPattern ( tree.getSubtreeWildcard ( ), { }, tree.getAlphabet ( ), tree.getContent ( ) ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( const RankedTree < SymbolType, RankType > & tree ) : PrefixRankedNonlinearPattern ( alphabet::SubtreeWildcardSymbol::instance < common::ranked_symbol < SymbolType, RankType > > ( ), { }, tree.getAlphabet ( ), toPrefixRanked ( tree.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( const RankedPattern < SymbolType, RankType > & tree ) : PrefixRankedNonlinearPattern ( tree.getSubtreeWildcard ( ), { }, tree.getAlphabet ( ), toPrefixRanked ( tree.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::PrefixRankedNonlinearPattern ( const RankedNonlinearPattern < SymbolType, RankType > & tree ) : PrefixRankedNonlinearPattern ( tree.getSubtreeWildcard ( ), tree.getNonlinearVariables ( ), tree.getAlphabet ( ), toPrefixRanked ( tree.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
std::vector < common::ranked_symbol < SymbolType, RankType > > PrefixRankedNonlinearPattern < SymbolType, RankType >::toPrefixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree ) {
	std::vector < common::ranked_symbol < SymbolType, RankType > > res;

	for ( typename ext::tree < common::ranked_symbol < SymbolType, RankType > >::const_prefix_iterator iter = tree.prefix_begin ( ); iter != tree.prefix_end ( ); ++iter )
		res.push_back ( * iter );

	return res;
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedNonlinearPattern < SymbolType, RankType >::clone ( ) const {
	return new PrefixRankedNonlinearPattern ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedNonlinearPattern < SymbolType, RankType >::plunder ( ) && {
	return new PrefixRankedNonlinearPattern ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const std::vector < common::ranked_symbol < SymbolType, RankType > > & PrefixRankedNonlinearPattern < SymbolType, RankType >::getContent ( ) const {
	return this->m_Data;
}

template < class SymbolType, class RankType >
void PrefixRankedNonlinearPattern < SymbolType, RankType >::setContent ( std::vector < common::ranked_symbol < SymbolType, RankType > > data ) {
	arityChecksum ( data );

	std::set < common::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < common::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet ( ).begin ( ), getAlphabet ( ).end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

template < class SymbolType, class RankType >
void PrefixRankedNonlinearPattern < SymbolType, RankType >::arityChecksum ( const std::vector < common::ranked_symbol < SymbolType, RankType > > & data ) {
	int arityChecksumAcc = 1;

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : data ) {
		arityChecksumAcc += ( unsigned ) symbol.getRank ( );
		arityChecksumAcc -= 1;
	}

	if ( arityChecksumAcc != 0 ) throw TreeException ( "The string does not form a tree" );
}

template < class SymbolType, class RankType >
bool PrefixRankedNonlinearPattern < SymbolType, RankType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType, class RankType >
int PrefixRankedNonlinearPattern < SymbolType, RankType >::compare ( const PrefixRankedNonlinearPattern & other ) const {
	auto first = std::tie ( m_Data, getAlphabet ( ), getSubtreeWildcard ( ), getNonlinearVariables ( ) );
	auto second = std::tie ( other.m_Data, other.getAlphabet ( ), other.getSubtreeWildcard ( ), other.getNonlinearVariables ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void PrefixRankedNonlinearPattern < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(PrefixRankedNonlinearPattern ";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType, class RankType >
PrefixRankedNonlinearPattern < SymbolType, RankType > PrefixRankedNonlinearPattern < SymbolType, RankType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedNonlinearPattern::getXmlTagName ( ) );
	common::ranked_symbol < SymbolType, RankType > subtreeWildcardSymbol = TreeFromXMLParser::parseSubtreeWildcardSymbol < common::ranked_symbol < SymbolType, RankType > > ( input );
	std::set < common::ranked_symbol < SymbolType, RankType > > nonlinearVariables = TreeFromXMLParser::parseNonlinearVariables < common::ranked_symbol < SymbolType, RankType > > ( input );
	std::set < common::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < common::ranked_symbol < SymbolType, RankType > > ( input );
	std::vector < common::ranked_symbol < SymbolType, RankType > > data = TreeFromXMLParser::parseLinearContent < common::ranked_symbol < SymbolType, RankType > > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedNonlinearPattern::getXmlTagName ( ) );

	return PrefixRankedNonlinearPattern < SymbolType, RankType > ( std::move ( subtreeWildcardSymbol ), std::move ( nonlinearVariables ), std::move ( rankedAlphabet ), std::move ( data ) );
}

template < class SymbolType, class RankType >
void PrefixRankedNonlinearPattern < SymbolType, RankType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedNonlinearPattern::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeSubtreeWildcard ( out, getSubtreeWildcard ( ) );
	TreeToXMLComposer::composeNonlinearVariables ( out, getNonlinearVariables ( ) );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedNonlinearPattern::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase* PrefixRankedNonlinearPattern < SymbolType, RankType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace alib {

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const std::vector < common::ranked_symbol < SymbolType, RankType > > & m_content = pattern.getContent ( );

		return std::find ( m_content.begin ( ), m_content.end ( ), symbol ) != m_content.end ( ) || pattern.template accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol;
	}

	static bool available ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::NonlinearAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return false;
	}

	static bool available ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return pattern.template accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol) {
		if ( ( unsigned ) symbol.getRank ( ) != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );

		if ( pattern.template accessElement < ::tree::SubtreeWildcard > ( ).get ( ) == symbol )
			throw ::tree::TreeException ( "Symbol " + ext::to_string ( symbol ) + "cannot be set as nonlinear variable since it is already subtree wildcard" );
	}
};

template < class SymbolType, class RankType >
class ElementConstraint< ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::SubtreeWildcard > {
public:
	static bool available ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		return pattern.template accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedNonlinearPattern < SymbolType, RankType > & pattern, const common::ranked_symbol < SymbolType, RankType > & symbol) {
		if ( ( unsigned ) symbol.getRank ( ) != 0 )
			throw ::tree::TreeException ( "SubtreeWildcard symbol has nonzero arity" );

		if ( pattern.template accessComponent < ::tree::NonlinearAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::tree::TreeException ( "Symbol " + ext::to_string ( symbol ) + "cannot be set as subtree wildcard since it is already nonlinear variable" );
	}
};

} /* namespace alib */

#endif /* PREFIX_RANKED_NONLINEAR_PATTERN_H_ */
