/*
 * PrefixRankedTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_TREE_H_
#define PREFIX_RANKED_TREE_H_

#include <set>
#include <vector>
#include <tree>
#include <sstream>
#include <algorithm>
#include <deque>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include <common/ranked_symbol.hpp>

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../RankedTreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include <tree/common/TreeNormalize.h>

namespace tree {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType, class RankType >
class PrefixRankedTree final : public RankedTreeBase, public alib::Components < PrefixRankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::vector < common::ranked_symbol < SymbolType, RankType > > m_Data;

	static std::vector < common::ranked_symbol < SymbolType, RankType > > toPrefixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree );

public:
	explicit PrefixRankedTree ( std::set < common::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedTree ( std::vector < common::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedTree ( const RankedTree < SymbolType, RankType > & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < common::ranked_symbol < SymbolType, RankType > > & getContent ( ) const;

	const std::set < common::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < common::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	void setContent ( std::vector < common::ranked_symbol < SymbolType, RankType > > data );

	void arityChecksum ( const std::vector < common::ranked_symbol < SymbolType, RankType > > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedTree & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "PrefixRankedTree";

		return xmlTagName;
	}

	static PrefixRankedTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef PrefixRankedTree < > normalized_type;

	virtual RankedTreeBase * normalize ( ) && {
		if ( typeid ( PrefixRankedTree < > ) == typeid ( PrefixRankedTree < SymbolType, RankType > ) )
			return this;

		std::set < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < common::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = alphabet::SymbolNormalize::normalizeRankedSymbols ( std::move ( m_Data ) );

		return new PrefixRankedTree < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace tree */

#include "RankedTree.h"

namespace tree {

template < class SymbolType, class RankType >
PrefixRankedTree < SymbolType, RankType >::PrefixRankedTree ( std::set < common::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < common::ranked_symbol < SymbolType, RankType > > data ) : alib::Components < PrefixRankedTree, common::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent ( std::move ( data ) );
}

template < class SymbolType, class RankType >
PrefixRankedTree < SymbolType, RankType >::PrefixRankedTree ( std::vector < common::ranked_symbol < SymbolType, RankType > > data ) : PrefixRankedTree ( std::set < common::ranked_symbol < SymbolType, RankType > > ( data.begin ( ), data.end ( ) ), data ) {
}

template < class SymbolType, class RankType >
PrefixRankedTree < SymbolType, RankType >::PrefixRankedTree ( const RankedTree < SymbolType, RankType > & tree ) : PrefixRankedTree ( tree.getAlphabet ( ), toPrefixRanked ( tree.getContent ( ) ) ) {
}

template < class SymbolType, class RankType >
std::vector < common::ranked_symbol < SymbolType, RankType > > PrefixRankedTree < SymbolType, RankType >::toPrefixRanked ( const ext::tree < common::ranked_symbol < SymbolType, RankType > > & tree ) {
	std::vector < common::ranked_symbol < SymbolType, RankType > > res;

	for ( typename ext::tree < common::ranked_symbol < SymbolType, RankType > >::const_prefix_iterator iter = tree.prefix_begin ( ); iter != tree.prefix_end ( ); ++iter )
		res.push_back ( * iter );

	return res;
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedTree < SymbolType, RankType >::clone ( ) const {
	return new PrefixRankedTree ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedTree < SymbolType, RankType >::plunder ( ) && {
	return new PrefixRankedTree ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const std::vector < common::ranked_symbol < SymbolType, RankType > > & PrefixRankedTree < SymbolType, RankType >::getContent ( ) const {
	return this->m_Data;
}

template < class SymbolType, class RankType >
void PrefixRankedTree < SymbolType, RankType >::setContent ( std::vector < common::ranked_symbol < SymbolType, RankType > > data ) {
	arityChecksum ( data );

	std::set < common::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < common::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

template < class SymbolType, class RankType >
void PrefixRankedTree < SymbolType, RankType >::arityChecksum ( const std::vector < common::ranked_symbol < SymbolType, RankType > > & data ) {
	int arityChecksumAcc = 1;

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : data ) {
		arityChecksumAcc += ( unsigned ) symbol.getRank ( );
		arityChecksumAcc -= 1;
	}

	if ( arityChecksumAcc != 0 ) throw TreeException ( "The string does not form a tree" );
}

template < class SymbolType, class RankType >
bool PrefixRankedTree < SymbolType, RankType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType, class RankType >
int PrefixRankedTree < SymbolType, RankType >::compare ( const PrefixRankedTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void PrefixRankedTree < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(PrefixRankedTree ";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType, class RankType >
PrefixRankedTree < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const common::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType, class RankType >
PrefixRankedTree < SymbolType, RankType > PrefixRankedTree < SymbolType, RankType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedTree::getXmlTagName() );
	std::set < common::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < common::ranked_symbol < SymbolType, RankType > > ( input );
	std::vector < common::ranked_symbol < SymbolType, RankType > > data = TreeFromXMLParser::parseLinearContent < common::ranked_symbol < SymbolType, RankType > > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedTree::getXmlTagName() );

	return PrefixRankedTree < SymbolType, RankType > ( std::move ( rankedAlphabet ), std::move ( data ) );
}

template < class SymbolType, class RankType >
void PrefixRankedTree < SymbolType, RankType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase* PrefixRankedTree < SymbolType, RankType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace alib {

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::PrefixRankedTree < SymbolType, RankType >, common::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedTree < SymbolType, RankType > & tree, const common::ranked_symbol < SymbolType, RankType > & symbol ) {
		const std::vector < common::ranked_symbol < SymbolType, RankType > > & content = tree.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
	}

	static bool available ( const ::tree::PrefixRankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedTree < SymbolType, RankType > &, const common::ranked_symbol < SymbolType, RankType > & ) {
	}
};

} /* namespace alib */

#endif /* PREFIX_RANKED_TREE_H_ */
