/*
 * DirectedGraph.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#include "DirectedGraph.h"
#include "AdjacencyListDirectedGraph.h"
#include "AdjacencyMatrixDirectedGraph.h"

#include <iostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/GraphFromXMLParser.h"
#include "../common/GraphToXMLComposer.h"
#include "../Graph.h"
#include "../undirected/UndirectedGraph.h"
#include <object/Object.h>
#include <core/xmlApi.hpp>

namespace graph {


DirectedGraph::DirectedGraph() {
}

void DirectedGraph::fromDirected( const DirectedGraph & orig )
{
	int value;
	// convert nodes and their values
	for (const Node & n : orig.getNodes()) {
		try {
			value = orig.getNodeValue( n );
			this->addNode( n, value );
		} catch (const std::out_of_range & ex) { // original node doesn't have value
			this->addNode( n );
		}
	}
	// convert edges and their values
	for (const DirectedEdge & e : orig.getEdges()) {
		try {
			value = orig.getEdgeValue( e );
			this->addEdge( DirectedEdge( e.getFromNode(), e.getToNode() ), value );
		} catch (const std::out_of_range & ex) { // original edge doesn't have value
			this->addEdge( DirectedEdge( e.getFromNode(), e.getToNode() ) );
		}
	}
	// convert embedding
	this->setEmbedding( orig.getEmbedding() );
}

void DirectedGraph::fromUndirected( const UndirectedGraph & orig )
{
	int value;
	// convert nodes and their values
	for (const Node & n : orig.getNodes()) {
		try {
			value = orig.getNodeValue( n );
			this->addNode( n, value );
		} catch (const std::out_of_range & ex) { // original node doesn't have value
			this->addNode( n );
		}
	}
	// convert edges and their values
	for (const UndirectedEdge & e : orig.getEdges()) {
		try {
			value = orig.getEdgeValue( e );
			this->addEdge( DirectedEdge( e.getFirstNode(), e.getSecondNode() ), value );
			this->addEdge( DirectedEdge( e.getSecondNode(), e.getFirstNode() ), value );
		} catch (const std::out_of_range & ex) { // original edge doesn't have value
			this->addEdge( DirectedEdge( e.getFirstNode(), e.getSecondNode() ) );
			this->addEdge( DirectedEdge( e.getSecondNode(), e.getFirstNode() ) );
		}
	}
	// convert embedding
	this->setEmbedding( orig.getEmbedding() );
}

GRAPH_TYPE DirectedGraph::getType() const
{
	return GRAPH_TYPE::DIRECTED;
}

bool DirectedGraph::addNode( const Node & node, int value )
{
	setNodeValue( node, value );
	return true;
}

bool DirectedGraph::removeNode( const Node & node )
{
	return nodeValues.erase( node ) == 1;
}

bool DirectedGraph::addEdge( const DirectedEdge & edge, int value )
{
	setEdgeValue( edge, value );
	return true;
}

bool DirectedGraph::removeEdge( const DirectedEdge & edge )
{
	return edgeValues.erase( edge ) == 1;
}

bool DirectedGraph::areNodesWeighted() const
{
	for (const Node & n : this->getNodes()) {
		if (nodeValues.find(n) == nodeValues.end())
			return false;
	}
	return true;
}

bool DirectedGraph::areEdgesWeighted() const
{
	for (const DirectedEdge & e : this->getEdges()) {
		if (edgeValues.find(e) == edgeValues.end())
			return false;
	}
	return true;
}

int DirectedGraph::getNodeValue( const Node & node ) const
{
	return nodeValues.at( node );
}

void DirectedGraph::setNodeValue( const Node & node, int value )
{
	nodeValues.insert( {node, value} );
}

int DirectedGraph::getEdgeValue( const DirectedEdge & edge ) const
{
	return edgeValues.at( edge );
}

void DirectedGraph::setEdgeValue( const DirectedEdge & edge, int value )
{
	edgeValues.insert( {edge, value} );
}

std::unordered_map< Node, std::vector< Node > > DirectedGraph::getEmbedding() const
{
	return m_embedding;
}

void DirectedGraph::setEmbedding( const std::unordered_map< Node, std::vector< Node > > embedding )
{
	this->m_embedding = embedding;
}

int DirectedGraph::compare( const DirectedGraph & other ) const
{
	std::set<Node> thisNodes = this->getNodes();
	std::set<DirectedEdge> thisEdges = this->getEdges();
	std::set<Node> otherNodes = other.getNodes();
	std::set<DirectedEdge> otherEdges = other.getEdges();

	auto first = std::tie(thisNodes, thisEdges, nodeValues, edgeValues, m_embedding);
	auto second = std::tie(otherNodes, otherEdges, other.nodeValues, other.edgeValues, other.m_embedding);

	static ext::compare<decltype(first)> comp;
	return comp(first, second);
}

void DirectedGraph::operator>>( std::ostream & out ) const
{
	out << "(DirectedGraph " << this->toString() << ")";
}

DirectedGraph::operator std::string() const
{
	return this->toString();
}

std::string DirectedGraph::toString() const
{
	std::stringstream ss;
	for (const Node &node : getNodes()) {
		ss << node;
	}
	for (const DirectedEdge &edge : getEdges()) {
		ss << edge;
	}
	return ss.str();
}
/*
DirectedGraph * DirectedGraph::parse(std::deque<sax::Token>::iterator& input)
{
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DirectedGraph::getXmlTagName());

	DirectedGraph * graph;
	switch (GraphFromXMLParser::parseRepresentation(input)) {
	 case REPRESENTATION::ADJACENCY_LIST:
		graph = new AdjacencyListDirectedGraph();
		break;
	 case REPRESENTATION::ADJACENCY_MATRIX:
		graph = new AdjacencyMatrixDirectedGraph();
		break;
	 case REPRESENTATION::INCIDENCE_MATRIX:

	 default:
		graph = NULL;
		break;
	}

	GraphFromXMLParser::parseNodes(input, *graph);
	GraphFromXMLParser::parseDirectedEdges(input, *graph);
	GraphFromXMLParser::parseNodeValues(input, *graph);
	GraphFromXMLParser::parseDirectedEdgeValues(input, *graph);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DirectedGraph::getXmlTagName());
	return graph;
}

void DirectedGraph::compose(std::deque<sax::Token>& out) const
{
	out.emplace_back(DirectedGraph::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	GraphToXMLComposer::composeRepresentation(out, representationToString(this->getRepresentation()));
	GraphToXMLComposer::composeNodes(out, this->getNodes());
	GraphToXMLComposer::composeEdges(out, this->getEdges());
	GraphToXMLComposer::composeNodeValues(out, *this);
	GraphToXMLComposer::composeEdgeValues(out, *this);

	out.emplace_back(DirectedGraph::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}
*/

} // namespace graph

/*
namespace alib {

auto directedGraphParserRegister = xmlApi<graph::Graph>::ParserRegister<graph::DirectedGraph>();
auto directedGraphParserRegister2 = xmlApi<alib::Object>::ParserRegister<graph::DirectedGraph>();

} // namespace alib
*/
