/*
 * UndirectedGraph.h
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 *  Modified by: Jan Broz
 */

#ifndef UNDIRECTED_GRAPH_H_
#define UNDIRECTED_GRAPH_H_

#include <unordered_map>
#include <vector>
#include <set>
#include <tuple>

#include "../GraphBase.h"
#include "../GraphTypes.h"
#include "../common/Macros.h"
#include "../common/Node.h"
#include "UndirectedEdge.h"

namespace graph {

class DirectedGraph;

class UndirectedGraph : public GraphBase {
	std::unordered_map< Node, int > nodeValues;
	std::unordered_map< UndirectedEdge, int > edgeValues;
	std::unordered_map< Node, std::vector< Node > > m_embedding;

public:
	UndirectedGraph();

	virtual GRAPH_TYPE getType() const override;
	virtual REPRESENTATION getRepresentation() const override = 0;

	/// constructs an undirected graph from any representation of other undirected graph
	void fromUndirected( const UndirectedGraph & orig );
	/// converts directed graph into undirected by removing edge direction
	void fromDirected( const DirectedGraph & orig );


	// interface for all graph reprezentations

	virtual std::set< Node > getNodes() const = 0;
	virtual std::set< UndirectedEdge > getEdges() const = 0;

	virtual size_t nodeCount() const = 0;
	virtual size_t edgeCount() const = 0;

	virtual bool hasNode( const Node & node ) const = 0;
	virtual bool hasEdge( const UndirectedEdge & edge ) const = 0;

	virtual bool hasEdge( const Node & from, const Node & to ) const = 0;
	virtual std::set< UndirectedEdge > findEdges( const Node & from, const Node & to ) const = 0;

	virtual std::set< Node > neighbors( const Node & node ) const = 0;
	virtual std::set< UndirectedEdge > neighborEdges( const Node & node ) const = 0;

	virtual bool addNode( const Node & node ) = 0;
	// these 2 methods must be called from the implementation in subclasses
	virtual bool addNode( const Node & node, int value ) = 0;
	virtual bool removeNode( const Node & node ) = 0;

	virtual bool addEdge( const UndirectedEdge & edge ) = 0;
	// these 2 methods must be called from the implementation in subclasses
	virtual bool addEdge( const UndirectedEdge & edge, int value ) = 0;
	virtual bool removeEdge( const UndirectedEdge & edge ) = 0;


	// weight of nodes and edges

	bool areNodesWeighted() const;
	bool areEdgesWeighted() const;

	int getNodeValue( const Node & node ) const;
	void setNodeValue( const Node & node, int value );

	int getEdgeValue( const UndirectedEdge & edge ) const;
	void setEdgeValue( const UndirectedEdge & edge, int value );

	std::unordered_map< Node, std::vector < Node > > getEmbedding() const;
	void setEmbedding( const std::unordered_map< Node, std::vector< Node > > embedding );


	// common methods

	int compare( const ObjectBase & other ) const override {
		if (std::type_index( typeid( *this ) ) == std::type_index( typeid( other ) ))
			return this->compare( (decltype( *this ))other );
		return std::type_index( typeid( *this ) ) - std::type_index( typeid( other ) );
	}
	int compare( const UndirectedGraph & other ) const;

	virtual void operator>>( std::ostream & out ) const override;
	virtual explicit operator std::string() const override;
	virtual std::string toString() const override;

	//static UndirectedGraph * parse( std::deque< sax::Token >::iterator & input );
	//void compose( std::deque< sax::Token > & out ) const override;

	/*static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UndirectedGraph";

		return xmlTagName;
	}*/

	friend class GraphToXMLComposer;
	friend class GraphToStringComposer;

};

} // namespace graph

#endif // UNDIRECTED_GRAPH_H_
