/*
 * Author: Radovan Cerveny
 */

#include "sax/FromXMLParserHelper.h"
#include "MeasurementProvisionerConfigurationXml.hpp"
#include "exception/CommonException.h"

namespace measurements {

MeasurementProvisionerConfiguration MeasurementProvisionerConfigurationXml::parse ( std::deque < sax::Token > & tokens ) {
	MeasurementProvisionerConfiguration configuration;

	std::deque < sax::Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	parseMeasurementProvisioner ( configuration, iter );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpected tokens at the end of the XML" );

	return configuration;
}

void MeasurementProvisionerConfigurationXml::parseMeasurementProvisioner ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == ENVIRONMENT_TAG )
			parseEnvironment ( mpc, iter );
		else if ( data == INPUT_DATA_TAG )
			parseInputData ( mpc, iter );
		else if ( data == PIPELINES_TAG )
			parsePipelines ( mpc, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_TAG );
}

void MeasurementProvisionerConfigurationXml::parseEnvironment ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, ENVIRONMENT_TAG );

	mpc.environment.workingDirectory = "";
	mpc.environment.pipelineIterations = 1;

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == WORKING_DIRECTORY_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, WORKING_DIRECTORY_TAG );
			mpc.environment.workingDirectory = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, WORKING_DIRECTORY_TAG );
		} else if ( data == PIPELINE_ITERATIONS_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINE_ITERATIONS_TAG );
			mpc.environment.pipelineIterations = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINE_ITERATIONS_TAG );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, ENVIRONMENT_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputData ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, INPUT_DATA_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == INPUT_DATA_BATCH_TAG )
			parseInputBatch ( mpc, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, INPUT_DATA_TAG );
}

void MeasurementProvisionerConfigurationXml::parseInputBatch ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, INPUT_DATA_BATCH_TAG );

	MPCInputBatch mpcib;

	mpcib.forcedDependencyBatch = false;

	if ( sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ATTRIBUTE ) ) {
		std::string attrName = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ATTRIBUTE );

		if ( attrName == INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ATTRIBUTE, attrName );
			std::string attrVal = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );

			if ( attrVal == "true" )
				mpcib.forcedDependencyBatch = true;
			else if ( attrVal == "false" )
				mpcib.forcedDependencyBatch = false;
			else
				throw::exception::CommonException ( "XML: MPCInputBatch malformed boolean attr: " + attrName );

			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ATTRIBUTE, attrName );
		}
	}

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		mpcib.batch.push_back ( parseInputDatum ( data, iter ) );
	}

	mpc.inputBatches.push_back ( std::move ( mpcib ) );

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, INPUT_DATA_BATCH_TAG );
}

MPCInputDatum MeasurementProvisionerConfigurationXml::parseInputDatum ( const std::string & tag, std::deque < sax::Token >::iterator & iter ) {
	MPCInputDatum mpcid;

	mpcid.id = -1;
	mpcid.count = 1;
	mpcid.alias = "";

	if ( tag == INPUT_DATA_FILE_TAG )
		mpcid.type = MPCInputDatum::Type::FILE;
	else if ( tag == INPUT_DATA_GENERATOR_TAG )
		mpcid.type = MPCInputDatum::Type::GENERATOR;
	else
		throw::exception::CommonException ( "XML: Unknown MPCInputDatum type" );

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, tag );

	bool foundId = false;

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ATTRIBUTE ) ) break;

		std::string attrName = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ATTRIBUTE );
		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ATTRIBUTE, attrName );
		std::string attrVal = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ATTRIBUTE, attrName );

		if ( attrName == INPUT_DATA_ATTR_ID_NAME ) {
			mpcid.id = ext::from_string < int > ( attrVal );
			foundId	 = true;
		} else if ( attrName == INPUT_DATA_ATTR_COUNT_NAME ) {
			mpcid.count = ext::from_string < int > ( attrVal );
		} else if ( attrName == INPUT_DATA_ATTR_ALIAS_NAME ) {
			mpcid.alias = std::move ( attrVal );
		} else {
			throw::exception::CommonException ( "XML: MPCInputDatum unknown attr: " + attrName );
		}
	}

	if ( !foundId )
		throw::exception::CommonException ( "XML: MPCInputDatum id attr was not specified" );

	mpcid.value = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, tag );
	return mpcid;
}

void MeasurementProvisionerConfigurationXml::parsePipelines ( MeasurementProvisionerConfiguration & mpc, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINES_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_TAG ) {
			mpc.pipelines.push_back ( { } );
			parsePipeline ( mpc.pipelines.back ( ), iter );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINES_TAG );
}

void MeasurementProvisionerConfigurationXml::parsePipeline ( MPCPipeline & pipeline, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINE_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_TAG )
			pipeline.push_back ( parsePipelineCommand ( iter ) );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINE_TAG );
}

MPCPipelineCommand MeasurementProvisionerConfigurationXml::parsePipelineCommand ( std::deque < sax::Token >::iterator & iter ) {

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COMMAND_TAG );

	MPCPipelineCommand mpcpc;

	mpcpc.measure = true;
	mpcpc.alias = "";

	if ( sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ATTRIBUTE ) ) {
		std::string attrName = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ATTRIBUTE );
		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ATTRIBUTE, attrName );
		std::string attrVal = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ATTRIBUTE, attrName );

		if ( attrName == COMMAND_ATTR_MEASURE_NAME ) {
			if ( attrVal == "true" )
				mpcpc.measure = true;
			else if ( attrVal == "false" )
				mpcpc.measure = false;
			else
				throw::exception::CommonException ( "XML: MPCPipelineCommand malformed boolean attr: " + attrName );
		} else if ( attrName == COMMAND_ATTR_ALIAS_NAME ) {
			mpcpc.alias = std::move ( attrVal );
		} else {
			throw::exception::CommonException ( "XML: MPCPipelineCommand unknown attr: " + attrName );
		}
	}

	mpcpc.command = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COMMAND_TAG );

	return mpcpc;
}

std::string MeasurementProvisionerConfigurationXml::MEASUREMENT_PROVISIONER_TAG = "MeasurementProvisioner";

std::string MeasurementProvisionerConfigurationXml::ENVIRONMENT_TAG = "Environment";
std::string MeasurementProvisionerConfigurationXml::WORKING_DIRECTORY_TAG = "WorkingDirectory";
std::string MeasurementProvisionerConfigurationXml::PIPELINE_ITERATIONS_TAG = "PipelineIterations";

std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_TAG = "InputData";
std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_TAG = "InputBatch";
std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_BATCH_ATTR_DEPENDENCY_NAME = "dependency";

std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_FILE_TAG = "File";
std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_GENERATOR_TAG = "Generator";

std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_ID_NAME = "id";
std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_COUNT_NAME = "count";
std::string MeasurementProvisionerConfigurationXml::INPUT_DATA_ATTR_ALIAS_NAME = "alias";

std::string MeasurementProvisionerConfigurationXml::PIPELINES_TAG = "Pipelines";
std::string MeasurementProvisionerConfigurationXml::PIPELINE_TAG	 = "Pipeline";
std::string MeasurementProvisionerConfigurationXml::COMMAND_TAG = "Command";

std::string MeasurementProvisionerConfigurationXml::COMMAND_ATTR_MEASURE_NAME = "measure";
std::string MeasurementProvisionerConfigurationXml::COMMAND_ATTR_ALIAS_NAME = "alias";
}
