/*
 * Author: Radovan Cerveny
 */

#include "MeasurementProvisionerResultsXml.hpp"
#include "exception/CommonException.h"
#include "sax/FromXMLParserHelper.h"
#include <measure>

namespace measurements {

std::deque < sax::Token > MeasurementProvisionerResultsXml::compose ( const MeasurementProvisionerResults & mpr ) {
	std::deque < sax::Token > tokens;

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULTS_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const MPRInputResult & mprir : mpr.inputResults )
		composeMeasurementProvisionerResult ( tokens, mprir );

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULTS_TAG, sax::Token::TokenType::END_ELEMENT );
	return tokens;
}

void MeasurementProvisionerResultsXml::composeMeasurementProvisionerResult ( std::deque < sax::Token > & tokens, const MPRInputResult & mprir ) {
	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULT_TAG, sax::Token::TokenType::START_ELEMENT );

	composeInputs ( tokens, mprir );
	composePipelines ( tokens, mprir );

	tokens.emplace_back ( MEASUREMENT_PROVISIONER_RESULT_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeInputs ( std::deque < sax::Token > & tokens, const MPRInputResult & mprir ) {
	tokens.emplace_back ( INPUTS_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const std::pair < const int, std::string > & input : mprir.inputs ) {
		tokens.emplace_back ( INPUT_TAG, sax::Token::TokenType::START_ELEMENT );
		tokens.emplace_back ( INPUT_ATTR_ID_NAME, sax::Token::TokenType::START_ATTRIBUTE );
		tokens.emplace_back ( ext::to_string ( input.first ), sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( INPUT_ATTR_ID_NAME, sax::Token::TokenType::END_ATTRIBUTE );
		tokens.emplace_back ( input.second, sax::Token::TokenType::CHARACTER );
		tokens.emplace_back ( INPUT_TAG, sax::Token::TokenType::END_ELEMENT );
	}

	tokens.emplace_back ( INPUTS_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipelines ( std::deque < sax::Token > & tokens, const MPRInputResult & mprir ) {
	tokens.emplace_back ( PIPELINES_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const MPRPipelineResult & mprpr : mprir.pipelineResults )
		composePipeline ( tokens, mprpr );

	tokens.emplace_back ( PIPELINES_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipeline ( std::deque < sax::Token > & tokens, const MPRPipelineResult & mprpr ) {
	tokens.emplace_back ( PIPELINE_TAG, sax::Token::TokenType::START_ELEMENT );

	composePipelineStatus ( tokens, mprpr );
	composeCommands ( tokens, mprpr );

	tokens.emplace_back ( PIPELINE_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composePipelineStatus ( std::deque < sax::Token > & tokens, const MPRPipelineResult & mprpr ) {

	tokens.emplace_back ( PIPELINE_STATUS_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( EXIT_CODE_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( ext::to_string ( mprpr.pipelineStatus.exitCode ), sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( EXIT_CODE_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( ERROR_ORIGIN_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpr.pipelineStatus.errorOrigin, sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( ERROR_ORIGIN_TAG, sax::Token::TokenType::END_ELEMENT );
	tokens.emplace_back ( ERROR_VALUE_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpr.pipelineStatus.errorValue, sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( ERROR_VALUE_TAG, sax::Token::TokenType::END_ELEMENT );

	tokens.emplace_back ( PIPELINE_STATUS_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeCommands ( std::deque < sax::Token > & tokens, const MPRPipelineResult & mprpr ) {
	tokens.emplace_back ( COMMANDS_TAG, sax::Token::TokenType::START_ELEMENT );

	for ( const MPRPipelineCommandResult & mprpcr : mprpr.commandResults )
		composeCommand ( tokens, mprpcr );

	tokens.emplace_back ( COMMANDS_TAG, sax::Token::TokenType::END_ELEMENT );
}

void MeasurementProvisionerResultsXml::composeCommand ( std::deque < sax::Token > & tokens, const MPRPipelineCommandResult & mprpcr ) {
	tokens.emplace_back ( COMMAND_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( COMMAND_NAME_TAG, sax::Token::TokenType::START_ELEMENT );
	tokens.emplace_back ( mprpcr.command, sax::Token::TokenType::CHARACTER );
	tokens.emplace_back ( COMMAND_NAME_TAG, sax::Token::TokenType::END_ELEMENT );

	measurements::MeasurementResultsXml::compose ( mprpcr.measurementResults, tokens );

	tokens.emplace_back ( COMMAND_TAG, sax::Token::TokenType::END_ELEMENT );
}

/*****************************************************************************************************/

MeasurementProvisionerResults MeasurementProvisionerResultsXml::parse ( std::deque < sax::Token > & tokens ) {
	MeasurementProvisionerResults mpr;

	std::deque < sax::Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw::exception::CommonException ( "Empty tokens list" );

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_RESULTS_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == MEASUREMENT_PROVISIONER_RESULT_TAG )
			mpr.inputResults.push_back ( parseMeasurementProvisionerResult ( iter ) );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_RESULTS_TAG );

	if ( iter != tokens.end ( ) ) throw::exception::CommonException ( "Unexpected tokens at the end of the XML" );

	return mpr;
}

MPRInputResult MeasurementProvisionerResultsXml::parseMeasurementProvisionerResult ( std::deque < sax::Token >::iterator & iter ) {
	MPRInputResult mprir;

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, MEASUREMENT_PROVISIONER_RESULT_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == INPUTS_TAG )
			parseInputs ( mprir, iter );
		else if ( data == PIPELINES_TAG )
			parsePipelines ( mprir, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, MEASUREMENT_PROVISIONER_RESULT_TAG );

	return mprir;
}

void MeasurementProvisionerResultsXml::parseInputs ( MPRInputResult & mprir, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, INPUTS_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, INPUT_TAG );

		std::string idAttr = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ATTRIBUTE );

		if ( idAttr != INPUT_ATTR_ID_NAME )
			throw::exception::CommonException ( "MPRInputResult Input expects only id attr" );

		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ATTRIBUTE, idAttr );

		int id = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );

		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ATTRIBUTE, idAttr );

		std::string inputName = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );

		mprir.inputs.emplace_back ( id, std::move ( inputName ) );

		sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, INPUT_TAG );
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, INPUTS_TAG );
}

void MeasurementProvisionerResultsXml::parsePipelines ( MPRInputResult & mprir, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINES_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_TAG )
			mprir.pipelineResults.push_back ( parsePipeline ( iter ) );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINES_TAG );
}

MPRPipelineResult MeasurementProvisionerResultsXml::parsePipeline ( std::deque < sax::Token >::iterator & iter ) {
	MPRPipelineResult mprpr;
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINE_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == PIPELINE_STATUS_TAG )
			parsePipelineStatus ( mprpr, iter );
		else if ( data == COMMANDS_TAG )
			parseCommands ( mprpr, iter );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINE_TAG );

	return mprpr;
}

void MeasurementProvisionerResultsXml::parsePipelineStatus ( MPRPipelineResult & mprpr, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, PIPELINE_STATUS_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == EXIT_CODE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, EXIT_CODE_TAG );
			mprpr.pipelineStatus.exitCode = ext::from_string < int > ( sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER ) );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, EXIT_CODE_TAG );
		} else if ( data == ERROR_ORIGIN_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, ERROR_ORIGIN_TAG );

			if ( sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::CHARACTER ) )
				mprpr.pipelineStatus.errorOrigin = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );

			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, ERROR_ORIGIN_TAG );
		} else if ( data == ERROR_VALUE_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, ERROR_VALUE_TAG );

			if ( sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::CHARACTER ) )
				mprpr.pipelineStatus.errorValue = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );

			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, ERROR_VALUE_TAG );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, PIPELINE_STATUS_TAG );
}

void MeasurementProvisionerResultsXml::parseCommands ( MPRPipelineResult & mprpr, std::deque < sax::Token >::iterator & iter ) {
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COMMANDS_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_TAG )
			mprpr.commandResults.push_back ( parseCommand ( iter ) );
		else
			break;
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COMMANDS_TAG );
}

MPRPipelineCommandResult MeasurementProvisionerResultsXml::parseCommand ( std::deque < sax::Token >::iterator & iter ) {
	MPRPipelineCommandResult mprpcr;
	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COMMAND_TAG );

	for ( ; ; ) {
		if ( !sax::FromXMLParserHelper::isTokenType ( iter, sax::Token::TokenType::START_ELEMENT ) ) break;

		std::string data = sax::FromXMLParserHelper::getTokenData ( iter, sax::Token::TokenType::START_ELEMENT );

		if ( data == COMMAND_NAME_TAG ) {
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::START_ELEMENT, COMMAND_NAME_TAG );
			mprpcr.command = sax::FromXMLParserHelper::popTokenData ( iter, sax::Token::TokenType::CHARACTER );
			sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COMMAND_NAME_TAG );
		} else if ( data == MEASUREMENT_RESULTS_TAG ) {
			mprpcr.measurementResults = measurements::MeasurementResultsXml::parse ( iter );
		} else {
			break;
		}
	}

	sax::FromXMLParserHelper::popToken ( iter, sax::Token::TokenType::END_ELEMENT, COMMAND_TAG );
	return mprpcr;
}

/*****************************************************************************************************/

std::string MeasurementProvisionerResultsXml::MEASUREMENT_PROVISIONER_RESULTS_TAG = "MeasurementProvisionerResults";

std::string MeasurementProvisionerResultsXml::MEASUREMENT_PROVISIONER_RESULT_TAG = "MeasurementProvisionerResult";

std::string MeasurementProvisionerResultsXml::INPUTS_TAG = "Inputs";

std::string MeasurementProvisionerResultsXml::INPUT_TAG = "Input";
std::string MeasurementProvisionerResultsXml::INPUT_ATTR_ID_NAME = "id";

std::string MeasurementProvisionerResultsXml::PIPELINES_TAG = "Pipelines";
std::string MeasurementProvisionerResultsXml::PIPELINE_TAG  = "Pipeline";

std::string MeasurementProvisionerResultsXml::PIPELINE_STATUS_TAG = "PipelineStatus";
std::string MeasurementProvisionerResultsXml::EXIT_CODE_TAG = "ExitCode";
std::string MeasurementProvisionerResultsXml::ERROR_ORIGIN_TAG = "ErrorOrigin";
std::string MeasurementProvisionerResultsXml::ERROR_VALUE_TAG  = "ErrorValue";

std::string MeasurementProvisionerResultsXml::COMMANDS_TAG = "Commands";
std::string MeasurementProvisionerResultsXml::COMMAND_TAG  = "Command";

std::string MeasurementProvisionerResultsXml::COMMAND_NAME_TAG = "Name";

std::string MeasurementProvisionerResultsXml::MEASUREMENT_RESULTS_TAG = "MeasurementResults";
}
