/*
 * AutomatonToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonToStringComposer.h"
#include "exception/CommonException.h"
#include "automaton/FSM/DFA.h"
#include "automaton/FSM/NFA.h"
#include "automaton/FSM/MultiInitialStateNFA.h"
#include "automaton/FSM/EpsilonNFA.h"
#include <registration/AlgoRegistration.hpp>

namespace automaton {

template < class SymbolType, class StateType >
void AutomatonToStringComposer::compose(std::ostream& out, const DFA < SymbolType, StateType >& automaton) {
	out << "DFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::ObjectToStringComposer::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::ObjectToStringComposer::compose(out, state);

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

template < class SymbolType, class StateType >
std::string AutomatonToStringComposer::compose ( const DFA < SymbolType, StateType > & automaton ) {
	std::stringstream ss;
	compose ( ss, automaton );
	return ss.str ( );
}

auto AutomatonToStringComposerDFA = registration::OverloadRegister < AutomatonToStringComposer, void, DFA < > >(AutomatonToStringComposer::compose);
auto AutomatonToStringComposerDFA2 = registration::AbstractRegister < AutomatonToStringComposer, std::string, const DFA < > & > ( AutomatonToStringComposer::compose );
auto ObjectToStringComposerDFA = registration::OverloadRegister < alib::ObjectBaseDispatcher, void, DFA < > >( AutomatonToStringComposer::compose);

template < class SymbolType, class StateType >
void AutomatonToStringComposer::compose(std::ostream& out, const NFA < SymbolType, StateType > & automaton) {
	out << "NFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::ObjectToStringComposer::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::ObjectToStringComposer::compose(out, state);

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

auto AutomatonToStringComposerNFA = registration::OverloadRegister < AutomatonToStringComposer, void, NFA < > >(AutomatonToStringComposer::compose);
auto ObjectToStringComposerNFA = registration::OverloadRegister < alib::ObjectBaseDispatcher, void, NFA < > >( AutomatonToStringComposer::compose);

template < class SymbolType, class StateType >
void AutomatonToStringComposer::compose(std::ostream& out, const MultiInitialStateNFA < SymbolType, StateType > & automaton) {
	out << "MISNFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::ObjectToStringComposer::compose(out, symbol);
	}

	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialStates().find(state) != automaton.getInitialStates().end()) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::ObjectToStringComposer::compose(out, state);

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

auto AutomatonToStringComposerMultiInitialStateNFA = registration::OverloadRegister < AutomatonToStringComposer, void, MultiInitialStateNFA < > >(AutomatonToStringComposer::compose);
auto ObjectToStringComposerMultiInitialStateNFA = registration::OverloadRegister < alib::ObjectBaseDispatcher, void, MultiInitialStateNFA < > >( AutomatonToStringComposer::compose);

template < class SymbolType, class EpsilonType, class StateType >
void AutomatonToStringComposer::compose(std::ostream& out, const EpsilonNFA < SymbolType, EpsilonType, StateType > & automaton) {
	out << "ENFA";
	for(const auto& symbol : automaton.getInputAlphabet()) {
		out << " ";
		alib::ObjectToStringComposer::compose(out, symbol);
	}

	out << " #E";
	out << std::endl;

	for(const auto& state : automaton.getStates()) {
		if(automaton.getInitialState() == state) {
			out << ">";
		}
		if(automaton.getFinalStates().find(state) != automaton.getFinalStates().end()) {
			out << "<";
		}
		alib::ObjectToStringComposer::compose(out, state);

		composeTransitionsFromState(out, automaton, state);

		out << std::endl;
	}
}

auto AutomatonToStringComposerEpsilonNFA = registration::OverloadRegister < AutomatonToStringComposer, void, EpsilonNFA < > >(AutomatonToStringComposer::compose);
auto ObjectToStringComposerEpsilonNFA = registration::OverloadRegister < alib::ObjectBaseDispatcher, void, EpsilonNFA < > >( AutomatonToStringComposer::compose);

void AutomatonToStringComposer::compose(std::ostream& output, const Automaton& automaton) {
	dispatch(output, automaton.getData());
}

auto ObjectToStringComposerAutomaton = registration::OverloadRegister < alib::WrapperBaseDispatcher, void, Automaton>(AutomatonToStringComposer::compose);

} /* namespace automaton */
