/*
 * arun.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <vector>
#include <sax/FromXMLParserHelper.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>
#include <string/String.h>
#include <primitive/Bool.h>
#include <object/Object.h>
#include <tree/ranked/RankedTree.h>
#include <automaton/Automaton.h>
#include <automaton/run/Accept.h>
#include <automaton/run/Result.h>
#include <automaton/run/Translate.h>
#include <automaton/run/Occurrences.h>
#include <container/ObjectsSet.h>

#include <iostream>

int main(int argc, char* argv[]) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd("Automaton run binary", ' ', "0.01");

		std::vector<std::string> allowed;
		allowed.push_back("occurrences");
		allowed.push_back("accept");
		allowed.push_back("result");
		allowed.push_back("translate");
		TCLAP::ValuesConstraint<std::string> allowedVals( allowed );

		TCLAP::ValueArg<std::string> type(	"t",	"type",		"Run type",		false,	"accept",	&allowedVals);
		cmd.add( type );

		TCLAP::ValueArg<std::string> input(	"i",	"input",	"Input from file",	false,	"-",		"file");
		cmd.add( input );

		TCLAP::ValueArg<std::string> failState(	"f",	"fail_state",	"Override default fail state",	false,	"-",	"file");
		cmd.add( failState );

		TCLAP::ValueArg<std::string> automaton(	"a",	"automaton",	"Automaton to run",	false,	"-",		"file");
		cmd.add( automaton );

		TCLAP::SwitchArg measure(		"m",	"measure",	"Measure times",		false);
		cmd.add( measure );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		measurements::start("Overal", measurements::Type::OVERALL);
		measurements::start("Input read", measurements::Type::AUXILIARY);

		alib::Object inputData = alib::XmlDataFactory::fromTokens (sax::FromXMLParserHelper::parseInput(input));
		automaton::Automaton automatonData = alib::XmlDataFactory::fromTokens (sax::FromXMLParserHelper::parseInput(automaton));

		measurements::end();
		measurements::start("Algorithm", measurements::Type::MAIN);

		if( type.getValue() == "occurrences") {
			std::set<unsigned> res = automaton::run::Occurrences::occurrences(automatonData, inputData);

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "accept") {
			bool res = automaton::run::Accept::accept(automatonData, inputData);

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "result" && failState.isSet ( ) ) {
			DefaultStateType failStateLabel = alib::XmlDataFactory::fromTokens (sax::FromXMLParserHelper::parseInput(failState));
			DefaultStateType res = automaton::run::Result::result(automatonData, inputData, failStateLabel);

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "result" ) {
			DefaultStateType res = automaton::run::Result::result(automatonData, inputData);

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else if( type.getValue() == "translate") {
			std::set < string::LinearString < > > res = automaton::run::Translate::translate(automatonData, inputData);

			measurements::end();
			measurements::start("Output write", measurements::Type::AUXILIARY);

			alib::XmlDataFactory::toStdout( res );
		} else {
			throw exception::CommonException( "Invalid run type" );
		}

		measurements::end();
		measurements::end();

		if(measure.getValue()) ext::cmeasure << measurements::results() << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout( exception );
		return 1;
	} catch (const TCLAP::ArgException& exception) {
		std::cout << exception.error() << std::endl;
		return 2;
	} catch (const std::exception& exception) {
		std::cerr << "Exception caught: " << exception.what() << std::endl;
		return 3;
	} catch (...) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
